# -*- coding: utf-8 -*-

import numpy as np
import os

from .properties import global_options, parameter_options, variable_options
from .gk_operators import GK_Operators

#%% Write files

def _build_model(self):
    ''' Write model to apm file.

    Also does some minimal model validation

    Returns:
        Does not return
    '''
    model = ''

    if self._constants:
        model += 'Constants\n'
        for const in self._constants:
            model += '\t%s = %s\n' % (const, const.value)
        model += 'End Constants\n'

    if self._parameters:
        model += 'Parameters\n'
        for parameter in self._parameters:
            i = 0
            model += '\t%s' % parameter
            if not isinstance(parameter.VALUE.value, (list,np.ndarray)):
                if not (parameter.VALUE==None):
                    i = 1
                    model += ' = %s' % parameter.VALUE
            if parameter.UPPER is not None:
                if i == 1:
                    model += ', '
                i = 1
                model += '<= %s' % parameter.UPPER
            if parameter.LOWER is not None:
                if i == 1:
                    model += ', '
                i = 1
                model += '>= %s' % parameter.LOWER
            model += '\n'
        model += 'End Parameters\n'

    if self._variables:
        model += 'Variables\n'
        for variable in self._variables:
            i = 0
            model += '\t%s' % variable
            if not isinstance(variable.VALUE.value, (list,np.ndarray)):
                if not (variable.VALUE==None):
                    i = 1
                    model += ' = %s' % variable.VALUE
            if variable.UPPER is not None:
                if i == 1:
                    model += ', '
                i = 1
                model += '<= %s' % variable.UPPER
            if variable.LOWER is not None:
                if i == 1:
                    model += ', '
                i = 1
                model += '>= %s' % variable.LOWER
            model += '\n'
        model += 'End Variables\n'

    if self._intermediates:
        model += 'Intermediates\n'
        for i in range(len(self._inter_equations)):
            model += '\t%s=%s\n' % (str(self._intermediates[i]), str(self._inter_equations[i]))
        model += 'End Intermediates\n'

    if self._equations or self._objectives:
        model += 'Equations\n'
        if self._equations:
            for equation in self._equations:
                model += '\t%s\n' % equation
        if self._objectives:
            for o in self._objectives:
                model += '\t%s\n' % o
        model += 'End Equations\n'

    if self._connections:
        model += 'Connections\n'
        for connection in self._connections:
            model += '\t%s\n' % connection
        model += 'End Connections\n'

    if self._objects:
        model += 'Objects\n'
        for obj_str in self._objects:
            model += '\t%s\n' % obj_str
        model += 'End Objects\n'

    if self._compounds:
        model += 'Compounds\n'
        for compound in self._compounds:
            model += '  %s\n' % (compound,)
        model += 'End Compounds\n'

    #print(model) #for debugging

    #replace multiple operators resulting from signs
    model = model.replace('++','+').replace('--','+').replace('+-','-').replace('-+','-')

    # Create .apm file
    if(self._model_name == None):
        self._model_name = "default_model_name"
    filename = self._model_name + '.apm'

    # Create file in writable format always overrite previous model file
    f = open(os.path.join(self._path,filename), 'w')
    f.write('Model\n')
    f.write(model)
    f.write('\nEnd Model')
    if self._raw:
        f.write('\n')
        for r in self._raw:
            f.write('%s\n'%r)
    f.close()

    self._model = 'auto-generated' #what does this do?

    self._model_initialized = True



def _write_csv(self):
    """Write csv file and validate data.
    If the problem is dynamic, the time discretization is provided in the
    first column of this csv. All params/variables that are initialized
    with an array are loaded as well and must be the same length. """

    file_name = self._model_name + '.csv'

    ## Dynamic data csv
    if self.options.IMODE > 3:
        #Start with time
        length = np.size(self.time)
        csv_data = np.hstack(('time',np.array(self.time).flatten().astype(object)))
        first_array = True
    ## SS data
    else:
        first_array = False
        if self.time is not None:
            print("Warning: model time only used for dynamic modes (IMODE>3)")

    #check all parameters and arrays
    for vp in self._variables+self._parameters:
        #Only save csv data if the user changed the value (changes registered in vp.value.change)
        if vp.value.change is False:
            continue
        else:
            if first_array == False:
                length = np.size(np.array(vp.value).flatten())
                if self.options.IMODE in (1,3) and length > 1:
                    raise Exception('This steady-state IMODE only allows scalar values.')
                elif self.options.IMODE == 2 and length == 1:
                    #in MPU, the first vp checked could be a scalar value (FV, param, var initial guess)
                    #but the CV is likely longer so skip scalar values (they will be set in the APM file)
                    continue


            if vp.value.change is True: #Save the entire array of values
                #skip variable if its value is a string (ie symbolic initialization)
                if isinstance(vp.VALUE.value,GK_Operators):
                    #reset change indicator
                    vp.value.change = False
                    #do nothing else, go to next variable
                    continue

                #discretize all values to arrays
                if not isinstance(vp.VALUE.value, (list,np.ndarray)):
                    vp.VALUE = np.ones(length)*vp.VALUE
                elif len(vp.VALUE) == 1:
                    vp.VALUE = np.ones(length)*vp.VALUE[0]
                #confirm that previously discretized values are the right length
                elif np.size(vp.VALUE.value) != length:
                    raise Exception('Data arrays must have the same length, and match time discretization in dynamic problems')
                #group data with column header
                t = np.hstack((vp.name,np.array(vp.VALUE.value).flatten().astype(object)))

            elif isinstance(vp.value.change,list): #only certain elements should be saved
                if not isinstance(vp.VALUE.value, (list,np.ndarray)):
                    vp.VALUE.value = np.ones(length)*vp.VALUE.value
                elif len(vp.VALUE) == 1:
                    vp.VALUE = np.ones(length)*vp.VALUE[0]
                t = np.array(vp.VALUE).astype(object)
                t[:] = ' '
                t[vp.value.change] = np.array(vp.value)[vp.value.change]
                t = np.hstack((str(vp),t.flatten().astype(object)))

            else: #somebody broke value.change
                raise Exception('Variable value modification monitor malfunction.')

            #reset change indicator
            vp.value.change = False

            #if a measurement exists, save a nonnumeric in
            #value array to allow measurement to be read in
            if hasattr(vp,'MEAS'):
                if vp.MEAS != None:
                    #vp.VALUE = np.array(vp.VALUE).astype(object)
                    if self.options.IMODE in [5,8] and vp.type=='CV':
                        #measurements in estimation go at the end of the horizon
                        #FDELAY shifts the location of the measurement
                        t[-1-vp.FDELAY] = 'measurement'
                    else:
                        t[1] = "measurement"

                    #reset MEAS so it doesn't get repeated on next solve
                    vp.MEAS = None

            #If a value was specified through a connection, ensure consistency in the
            #csv file, otherwise the requested specified value will be overridden by
            #whatever initialization value is in the csv
            if hasattr(vp,'_override_csv'):
                for i in vp._override_csv: #for each tuple of (position,value)
                    #set value in t array
                    t[i[0]+1] = i[1] #index is +1 because of prepended header

            if first_array == False:
                csv_data = t
                first_array = True
            else:
                try:
                    csv_data = np.vstack((csv_data,t))
                except ValueError:
                    raise Exception('All variable value arrays must be the same length (and match the length of model time in dynamic problems).')

    #print(csv_data)
    #save array to csv
    if first_array == False: #no data
        self.csv_status = 'none'
    else:
        np.savetxt(os.path.join(self._path,file_name), csv_data.T, delimiter=",", fmt='%1.25s')
        self.csv_status = 'generated'



def _write_info(self):
    #since there is currently no way to change variable classification after
    #declaration, rewriting the info file is redundant and makes the server info
    #file on remote solves large
    #avoid this problem by only writing the info file for the first successful solve
    if self.options.CYCLECOUNT < 1:
        #Classify variable in .info file
        filename = self._model_name+'.info'

        #Create and open configuration files
        with open(os.path.join(self._path,filename), 'w+') as f:
            #check each Var and Param for FV/MV/SV/CV
            for vp in self._variables+self._parameters:
                if vp.type is not None:
                    f.write(vp.type+', '+vp.name+'\n')


def _generate_dbs_file(self):
    '''Write options to measurements.dbs file so it gets automatically deleted
    to prevent file build-up on the server

    Returns:
        Does not return
    '''
    #set filename
    filename = 'measurements.dbs'
    #print all global options
    file_content = self.options.getOverridesString()
    #cycle through all Params and Vars to find set options
    with open(os.path.join(self._path,filename), 'w+') as f:
        f.write(file_content)
        #check for set options of each Var and Param
        for vp in self._parameters:
            for o in parameter_options[vp.type]['inputs']+parameter_options[vp.type]['inout']:
                if o == 'VALUE':
                    continue
                else: #everything else is an option
                    if vp.__dict__[o] is not None:
                        f.write(vp.name+'.'+o+' = '+str(vp.__dict__[o])+'\n')

        for vp in self._variables:
            for o in variable_options[vp.type]['inputs']+variable_options[vp.type]['inout']:
                if o == 'VALUE':
                    continue
                else: #everything else is an option
                    if vp.__dict__[o] is not None:
                        f.write(vp.name+'.'+o+' = '+str(vp.__dict__[o])+'\n')


def _write_solver_options(self):
    opt_file = ''
    if self.solver_options:
        #determine filename from solver number
        if self.options.SOLVER == 1:
            filename = 'apopt.opt'
        elif self.options.SOLVER == 3:
            filename = 'ipopt.opt'
        else:
            raise TypeError("Solver options only available for APOPT(1) and IPOPT(3)")

        #write each option to a line
        for option in self.solver_options:
            opt_file += option + '\n'

        #If remote solve, pass string to append to .apm file
        if self._remote is True:
            return 'File ' + filename + '\n' + opt_file + 'End File\n'
        #write file for local solve
        else:
            with open(os.path.join(self._path,filename), 'w+') as f:
                f.write(opt_file)

    #do nothing if no options were added
    else:
        return opt_file

    opt_file += 'End File\n'
    return opt_file


#%% Not currently used


def _jsonify(self,data): #This function was mostly copied from SO
    if type(data).__module__=='numpy': # if value is numpy.*: > to python list
        json_data = data.tolist()
    elif isinstance(data, dict): # for nested lists
        json_data = dict()
        for key, value in data.iteritems():
            if isinstance(value, list): # for lists
                value = [ self.jsonify(item) if isinstance(item, dict) else item for item in value ]
            if isinstance(value, dict): # for nested lists
                value = self.jsonify(value)
            if isinstance(key, int): # if key is integer: > to string
                key = str(key)
            if type(value).__module__=='numpy': # if value is numpy.*: > to python list
                value = value.tolist()
            json_data[key] = value
    else:
        json_data = data
    return json_data


def _to_JSON(self): #JSON input to APM not currently supported -- this function isn't tested
    """
    include in JSON:
    global options
    variables (const,param,inter,var)
        name
        value
        type
        option_list
    """
    json_data = dict()
    #global options
    o_dict = dict()
    for o in global_options['inputs']+global_options['inout']:
        o_dict[o] = getattr(self.options,o)
    json_data['global options'] = o_dict

    if self.time is not None:
        json_data['time'] = self.jsonify(self.time)
    #Constants can't and won't change so there's no reason to pass them in the JSON
    #constant values must be given in the model file. Changing constant values requires recompiling the model.
#        #constants
#        if self._constants:
#            const_dict = dict()
#            for const in self._constants:
#                const_dict['value'] = self.jsonify(const.value)
#            const_dict[const.name] = const_dict

    if self._parameters:
        p_dict = dict()
        for parameter in self._parameters:
            o_dict = dict()
            for o in parameter_options[parameter.type]['inputs']+parameter_options[parameter.type]['inout']:
                if o == 'VALUE':
                    o_dict['VALUE'] = self.jsonify(parameter.value)
                else:
                    o_dict[o] = getattr(parameter,o)
            o_dict['type'] = parameter.type
            p_dict[parameter.name] = o_dict
        json_data['parameters'] = p_dict

    if self._variables:
        p_dict = dict()
        for parameter in self._variables:
            o_dict = dict()
            for o in variable_options[parameter.type]['inputs']+variable_options[parameter.type]['inout']:
                if o == 'VALUE':
                    o_dict['VALUE'] = self.jsonify(parameter.value)
                else:
                    o_dict[o] = getattr(parameter,o)
            o_dict['type'] = parameter.type
            p_dict[parameter.name] = o_dict
        json_data['variables'] = p_dict

    """
    if self._intermediates:
        temp_dict = dict()
        for intermediate in self._intermediates:
            temp_dict['name'] = {'value':self.jsonify(intermediate.value)}
        json_data['intermediates'] = temp_dict
    """
    f = open(os.path.join(self._path,'jsontest.json'), 'w')
    #f.write(json.dumps(self, default=lambda o: _try(o), sort_keys=True, indent=2, separators=(',',':')).replace('\n', ''))
    json.dump(json_data,f, indent=2,separators=(',', ':'))
    f.close()
    #return json.dumps(self, default=lambda o: _try(o), sort_keys=True, indent=0, separators=(',',':')).replace('\n', '')
    #load JSON to dictionary:
    #with open(os.path.join(self._path,'jsontest.json')) as json_file:
    #   data = json.load(json_file)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import json

from .properties import parameter_options, variable_options


#%% Look for logic gotchas

def gk_logic_tree(self):
    #%% Global options
    if self.options.CSV_READ == 1:
        print("Warning: Variables will not be initialized with CSV_READ=1")
    elif self.options.CSV_READ == 0:
        print("Warning: Non-scalar Variable and Parameter values will not be used with CSV_READ=0" )
        
    if self.options.CYCLECOUNT > 1 and self.options.TIME_SHIFT != 0:
        print("Warning: All variables were time shifted on subsequent resolves")
        
    #%% Control
    if self.options.IMODE % 3 == 0: #3,6,9
        
        #SP vs SPHI/SPLO
        if self.options.CV_TYPE == 1: # 1-norm of cv setpoint
            #CV SPHI and SPLO should be used, no SP
            for v in self._variables:
                if v.type == 'CV':
                    if v.STATUS == 1:
                        if v.SP is not None:
                            print("Warning: Use SPHI and SPLO (not SP) when EV_TYPE is 1")
                            break #teaching completed; don't need to check any more
                        if v.WSP is not None:
                            print("Warning: Use WSPHI and WSPLO (not WSP) when EV_TYPE is 1")
                            break #teaching completed; don't need to check any more
        elif self.options.CV_TYPE == 2: # 2-norm of cv setpoint
            #CV SPHI and SPLO should be used, no SP
            for v in self._variables:
                if v.type == 'CV':
                    if v.STATUS == 1:
                        if v.SPHI is not None or v.SPLO is not None:
                            print("Warning: Use SP (not SPHI and SPLO) when EV_TYPE is 2")
                            break #teaching completed; don't need to check any more
                        if v.WSPHI is not None or v.WSPLO is not None:
                            print("Warning: Use WSP (not WSPHI and WSPLO) when EV_TYPE is 2")
                            break #teaching completed; don't need to check any more
        
        if self.options.IMODE > 5: # 6 or 9 -- dynamic control
            for v in self._variables:
                if v.type == 'CV':
                    if v.STATUS == 1:
                        if v.TR_INIT > 0:
                            if v.TAU > self.time[-1]:
                                print("Warning: Trajectory time constant (TAU) longer than time horizon and TR_INIT not 0")
                        if v.TR_OPEN is not None:
                            if self.options.CV_TYPE != 1:
                                print("Warning: TR_OPEN only for CV_TYPE=1")
                    
    #%% Estimation
    if (self.options.IMODE+1)%3 == 0: #2,5,8
        for v in self._variables:
            if v.type == 'CV':
                if v.STATUS == 1:
                    print("Warning: STATUS of CVs has no effect in estimation. Use FSTATUS.")
                if v.MEAS_GAP is not None and self.options.EV_TYPE != 1:
                    print("Warning: MEAS_GAP only for EV_TYPE=1")
    
    #%% Measurements
    for vp in self._variables+self._parameters:
        if vp.type is not None:
            if vp.MEAS is not None:
                if vp.FSTATUS == 0:
                    print("Warning: MEAS not used when FSTATUS = 0")
                    
                    
                    

#%% Compare GEKKO and APM


#define comparison of options between APM and GEKKO
#to avoid false positive when floats are off by a little
def like(self,one,two):
    if isinstance(one,str): #string are compared directly
        return one==two
    else:
        return (one+self.options.OTOL >= two and one-self.options.OTOL <= two)

def verify_input_options(self):
    ## Load data
    f = open(os.path.join(self.path,'options.json'))
    data = json.load(f)
    f.close()
    ## Global Options
    for o in self.options._input_option_list: #for each global input option
        if o == 'CSV_READ' and self.csv_status == 'none':
            continue
        if self.options.__dict__[o] != data['APM'][o]: #compare APM to GK
            print(str(o)+" was not written correctly") #give message if they don't match
    ## Local Options
    for vp in self._parameters:
        if vp.type != None: #(FV/MV/SV/CV) not Param or Var
            for o in parameter_options[vp.type]['inputs']:
                if o not in ['LB','UB']: #TODO: for o in data[vp.name] to avoid this check
                    if vp.__dict__[o] is not None and not self.like(vp.__dict__[o], data[vp.name][o]):
                        print(str(vp)+'.'+str(o)+" was not written correctly") #give message if they don't match

    for vp in self._variables:
        if vp.type != None: #(FV/MV/SV/CV) not Param or Var
            for o in variable_options[vp.type]['inputs']:
                if o not in ['LB','UB']:
                    if vp.__dict__[o] is not None and not self.like(vp.__dict__[o], data[vp.name][o]):
                        print(str(vp)+'.'+str(o)+" was not written correctly") #give message if they don't match
                        
#%% Name Check
# Default GEKKO names are unique and valid by nature of their assignment. 
# User-defined names may not be unique or valid. Since the use must deliberately set names,
# this function exists of them to check but is not run by default.

def name_check(self):
    all_names = set() #build the set of all variables names to check for uniqueness
    #illegal names (reserved for functions)
    # not checking for 5+ character object names
    illegal = set(['abs','arx','cos','erf','exp','log','pwl',\
                     'sin','sum','tan','abs2','abs3','max2',\
                     'max3','min2','min3','sqrt','asin','acos',\
                     'atan','vsum'])
    
    for x in self._constants+self._parameters+self._variables+self._intermediates+self._objects:
        #unique names
        if x.name not in all_names:
            all_names.add(x.name)
        else:
            raise NameError(x.name+" is used multiple times")
        #reserved functions
        if len(x.name) >=3:
            if x.name in illegal: 
                raise NameError(x.name+" is an illegal name (reserved function name)")
        
        #special names
        if len(x.name) >=3:
            if x.name[0:3] == 'obj':
                print("Warning: "+x.name+" is added to the objective function (name starts with 'obj')")
            if x.name[0:3] == 'slk':
                print("Warning: "+x.name+" is a slack variable (name starts with 'slk') -- lower bound of zero")



# -*- coding: utf-8 -*-
from gekko import GEKKO
import numpy as np
from .gk_parameter import GKParameter
from .gk_variable import GKVariable
"""
GEKKO specializes in a optimization and control. This module extends GEKKO with 
chemical compounds, thermodynamic properties, and flowsheet objects.
"""

class Properties():
    
    def __init__(self,m):
        self.m = m
        # True if thermo object is created
        # Enforces compound definition before thermo_obj definition
        self._thermo_obj = False
        
    def compound(self,name):
        """ Add chemical compound to model with one of the following:
        1. IUPAC Name  (1,2-ethanediol)
        2. Common Name (ethylene glycol) 
        3. CAS Number (107-21-1)
        4. Formula (C2H6O2)
        Repeated compounds are permitted. All compounds should be declared
          before thermo objects are created. An error message will occur if
          the compound is not in the database and a file 'compounds.txt' will
          be created to communicate the available compounds.
        """        
        #verify that compound is not added after thermo objects
        if self._thermo_obj:
            raise TypeError("Define compound ("+name+") before creating a thermo object")
        # add compound name
        self.m._compounds.append(name)
        return
                
    def thermo(self,prop,T=300.0):
        """ Thermodynamic Properties
          usage: thermo('mw') for constants
                 thermo('lvp',T) for temperature dependent
        ---- Temperature Independent ----
        mw   = Molecular Weight (kg/kmol)
        tc   = Critical Temperature (K)
        pc   = Critical Pressure (Pa)
        vc   = Critical Volume (m^3/kmol)
        ccf  = Crit Compress Factor (unitless)
        mp   = Melting Point (K)
        tpt  = Triple Pt Temperature (K)
        tpp  = Triple Pt Pressure (Pa)
        nbp  = Normal Boiling Point (K)
        lmv  = Liq Molar Volume (m^3/kmol)
        ighf = IG Heat of Formation (J/kmol)
        iggf = IG Gibbs of Formation (J/kmol)
        igae = IG Absolute Entropy (J/kmol*K)
        shf  = Std Heat of Formation (J/kmol)
        sgf  = Std Gibbs of Formation (J/kmol)
        sae  = Std Absolute Entropy (J/kmol*K)
        hfmp = Heat Fusion at Melt Pt (J/kmol)
        snhc = Std Net Heat of Comb (J/kmol)
        af   = Acentric Factor (unitless)
        rg   = Radius of Gyration (m)
        sp   = Solubility Parameter ((J/m^3)^0.5)
        dm   = Dipole Moment (c*m)
        r    = van der Waals Volume (m^3/kmol)
        q    = van der Waals Area (m^2)
        ri   = Refractive Index (unitless)
        fp   = Flash Point (K)
        lfl  = Lower Flammability Limit (K)
        ufl  = Upper Flammability Limit (K)
        lflt = Lower Flamm Limit Temp (K)
        uflt = Upper Flamm Limit Temp (K)
        ait  = Auto Ignition Temp (K)
        ---- Temperature Dependent ----   
        sd   = Solid Density (kmol/m^3)
        ld   = Liquid Density (kmol/m^3) 
        svp  = Solid Vapor Pressure (Pa) 
        lvp  = Liquid Vapor Pressure (Pa) 
        hvap = Heat of Vaporization (J/kmol) 
        scp  = Solid Heat Capacity (J/kmol*K) 
        lcp  = Liquid Heat Capacity (J/kmol*K) 
        igcp = Ideal Gas Heat Capacity (J/kmol*K) 
        svc  = Second Virial Coefficient (m^3/kmol) 
        lv   = Liquid Viscosity (Pa*s) 
        vv   = Vapor Viscosity (Pa*s) 
        sk   = Solid Thermal Conductivity (W/m*K) 
        lk   = Liq Thermal Conductivity (W/m*K) 
        vk   = Vap Thermal Conductivity (W/m*K) 
        st   = Surface Tension (N/m) 
        sh   = Solid Enthalpy (J/kmol) 
        lh   = Liq Enthalpy (J/kmol) 
        vh   = Vap Enthalpy (J/kmol)                  
        """
        self._thermo_obj = True
        prop = prop.lower()
        
        # check if it is a temperature dependent property
        tdp = ['sd','ld','lv','vv','sk','lk','vk','st','sh','lh','vh',\
               'svp','lvp','scp','lcp','svc','hvap','igcp']
        if prop.lower() in tdp:
            td = True
            # inquire if T is a valid GEKKO variable or parameter
            if isinstance(T,(GKVariable,GKParameter)):
                Tin = T
            else:
                # create input variable if it is an expression
                Tin = self.m.Var()
                self.m.Equation(Tin==T)
        else:
            td = False

        # build thermo object with unique object name
        thermo_name = 'thermo_' + str(len(self.m._objects) + 1)
        self.m._objects.append(thermo_name+'=thermo_'+prop)

        # add connections between y and thermo object attribute y
        if not td:  # not temperature dependent
            y = {}
            i = 0
            for c in self.m._compounds:
                i += 1
                y[c] = self.m.Param()
                self.m._connections.append(y[c].name+'='+thermo_name+'.'+prop+'['+str(i)+']')
        else:  # temperature dependent
            y = {}
            i = 0
            for c in self.m._compounds:
                i += 1
                y[c] = self.m.Var()
                self.m._connections.append(y[c].name+'='+thermo_name+'.'+prop+'['+str(i)+']')
            # link temperature
            y['T'] = Tin
            self.m._connections.append(Tin.name+'='+thermo_name+'.T')
            
        # add units and property description
        if (prop=='mw'): y['units']='kg/kmol'; y['property']='Molecular Weight'
        if (prop=='tc'): y['units']='K'; y['property']='Critical Temperature'
        if (prop=='pc'): y['units']='Pa'; y['property']='Critical Pressure'
        if (prop=='vc'): y['units']='m^3/kmol'; y['property']='Critical Volume'
        if (prop=='ccf'): y['units']='unitless'; y['property']='Crit Compress Factor'
        if (prop=='mp'): y['units']='K'; y['property']='Melting Point'
        if (prop=='tpt'): y['units']='K'; y['property']='Triple Pt Temperature'
        if (prop=='tpp'): y['units']='Pa'; y['property']='Triple Pt Pressure'
        if (prop=='nbp'): y['units']='K'; y['property']='Normal Boiling Point'
        if (prop=='lmv'): y['units']='m^3/kmol'; y['property']='Liq Molar Volume'
        if (prop=='ighf'): y['units']='J/kmol'; y['property']='IG Heat of Formation'
        if (prop=='iggf'): y['units']='J/kmol'; y['property']='IG Gibbs of Formation'
        if (prop=='igae'): y['units']='J/kmol-K'; y['property']='IG Absolute Entropy'
        if (prop=='shf'): y['units']='J/kmol'; y['property']='Std Heat of Formation'
        if (prop=='sgf'): y['units']='J/kmol'; y['property']='Std Gibbs of Formation'
        if (prop=='sae'): y['units']='J/kmol-K'; y['property']='Std Absolute Entropy'
        if (prop=='hfmp'): y['units']='J/kmol'; y['property']='Heat Fusion at Melt Pt'
        if (prop=='snhc'): y['units']='J/kmol'; y['property']='Std Net Heat of Comb'
        if (prop=='af'): y['units']='unitless'; y['property']='Acentric Factor'
        if (prop=='rg'): y['units']='m'; y['property']='Radius of Gyration'
        if (prop=='sp'): y['units']='(J/m^3)^0.5'; y['property']='Solubility Parameter'
        if (prop=='dm'): y['units']='c*m'; y['property']='Dipole Moment'
        if (prop=='r'): y['units']='m^3/kmol'; y['property']='van der Waals Volume'
        if (prop=='q'): y['units']='m^2'; y['property']='van der Waals Area'
        if (prop=='ri'): y['units']='unitless'; y['property']='Refractive Index'
        if (prop=='fp'): y['units']='K'; y['property']='Flash Point'
        if (prop=='lfl'): y['units']='K'; y['property']='Lower Flammability Limit'
        if (prop=='ufl'): y['units']='K'; y['property']='Upper Flammability Limit'
        if (prop=='lflt'): y['units']='K'; y['property']='Lower Flamm Limit Temp'
        if (prop=='uflt'): y['units']='K'; y['property']='Upper Flamm Limit Temp'
        if (prop=='ait'): y['units']='K'; y['property']='Auto Ignition Temp'
        if (prop=='sd'): y['units']='kmol/m^3'; y['property']='Solid Density'
        if (prop=='ld'): y['units']='kmol/m^3 '; y['property']='Liquid Density'
        if (prop=='svp'): y['units']='Pa '; y['property']='Solid Vapor Pressure'
        if (prop=='lvp'): y['units']='Pa '; y['property']='Liquid Vapor Pressure'
        if (prop=='hvap'): y['units']='J/kmol '; y['property']='Heat of Vaporization'
        if (prop=='scp'): y['units']='J/kmol-K '; y['property']='Solid Heat Capacity'
        if (prop=='lcp'): y['units']='J/kmol-K '; y['property']='Liquid Heat Capacity'
        if (prop=='igcp'): y['units']='J/kmol-K '; y['property']='Ideal Gas Heat Capacity'
        if (prop=='svc'): y['units']='m^3/kmol '; y['property']='Second Virial Coefficient'
        if (prop=='lv'): y['units']='Pa*s '; y['property']='Liquid Viscosity'
        if (prop=='vv'): y['units']='Pa*s '; y['property']='Vapor Viscosity'
        if (prop=='sk'): y['units']='W/m-K '; y['property']='Solid Thermal Conductivity'
        if (prop=='lk'): y['units']='W/m-K '; y['property']='Liq Thermal Conductivity'
        if (prop=='vk'): y['units']='W/m-K '; y['property']='Vap Thermal Conductivity'
        if (prop=='st'): y['units']='N/m '; y['property']='Surface Tension'
        if (prop=='sh'): y['units']='J/kmol '; y['property']='Solid Enthalpy'
        if (prop=='lh'): y['units']='J/kmol '; y['property']='Liq Enthalpy'
        if (prop=='vh'): y['units']='J/kmol '; y['property']='Vap Enthalpy'
        
        return y

class ReserveObj:
    '''Reserve Object
          P = Pressure (Pa)
          T = Temperature (K)
          n = Molar holdup (kmol)
          x = Array of mole fractions
          phase = Phase (solid, liquid, vapor)
    '''
    name = ''
    P = None
    T = None
    n = None
    x = []
    phase = None

class StreamObj:
    '''Stream Object
          P = Pressure (Pa)
          T = Temperature (K)
          ndot = Molar flow rate (kmol/sec)
          x = Array of mole fractions
          phase = Phase (solid, liquid, vapor)
    '''
    name = ''
    P = None
    T = None
    ndot = None
    x = []
    phase = None

class FlashObj:
    '''Flash Object
          P = Pressure (Pa)
          T = Temperature (K)
          Q = Heat input (J/sec)
          gamma = Activity coefficients
          n = Molar holdup in flash column (kmol)
          inlet = Inlet stream
          reserve = Molar holdup
          vapor = Vapor stream
          liquid = Liquid stream
    '''
    name = ''
    P = None
    T = None
    Q = None
    gamma = []
    n = None
    inlet = ''
    reserve = ''
    vapor = ''
    liquid = ''
    
class MassObj:
    '''Mass Object
          m = mass (kg)
          reserve = reserve name or object
    '''
    name = ''
    m = None
    reserve = ''

class MassflowObj:
    '''Massflow Object
          mdot = mass flow (kg/sec)
          mdoti = mass flows of components (kg/sec)
          stream = stream name or object
    '''
    name = ''
    mdot = None
    mdoti = []
    stream = ''
    
class MolarflowObj:
    '''Molarflow Object
          ndot = molar flow (kmol/sec)
          ndoti = molar flows of components (kmol/sec)
          stream = stream name or object
    '''
    name = ''
    ndot = None
    ndoti = []
    stream = ''

class MixerObj:
    '''Mixer Object
          inlet = list of inlet stream names or objects
          outlet = outlet stream name or object
    '''
    name = ''
    inlet = []
    outlet = ''
    
class PIDObj:
    '''PID (Proportional Integral Derivative) Object
          co = Controller Output (u)
          pv = Process Variable (y)
          sp = Process Variable Set Point (ysp)
          Kc = PID Proportional constant
          tauI = PID Integral constant
          tauD = PID Derivative constant
          i = Integral term  
       
       Description:  PID: Proportional Integral Derivative Controller
       In the frequency domain the PID controller is described by
        U(s) = Kc*Y(s) + Y(s)*Kc/s*tauI + Kc*taud*s*Y(s)
       In the time domain the PID controller is described by
        u(t) = Kc*(ysp-y(t)) + (Kc/taui)*Integral(t=0...t)(ysp-y(t))dt + Kc*taud*dy(t)/dt
       This implementation splits the single equation into two equations
       The second equation is necessary to avoid the numerical integration.
       The equations are posed in an open equation form.  The integral time
        constant is multiplied through to avoid potential divide by zero.
       This form may have an advantage over placing the term taui in with the
        integral equation for cases where taui becomes very small.
        0 = -u*taui + Kc*((ysp-y)*taui + Integral + taud*(dy/dt)*taui)
        0 = d(Integral)/dt - (ysp-y)          
    '''
    name = ''
    co = None
    pv = None
    sp = None
    Kc = None
    tauI = None
    tauD = None
    i = None
    
class PumpObj:
    '''Pump Object
          dp = change in pressure
          inlet = inlet stream name or object
          outlet = outlet stream name or object
    '''
    name = ''
    dp = None
    inlet = ''
    outlet = ''
   
class ReactorObj:
    '''Reactor Object
       The reactor model is equivalent to the vessel model but also
        includes generation of species through reaction.  The reaction rates
        are defined as (+) generation and (-) consumption.
       In addition to the reaction rates, there is a term for heat generation
        from exothermic reactions (+) or heat removal from endothermic reactions (-)
          V = Volume (m^3)
          Q = Heat input (J/sec)
          Qr = Heat generation by reaction (J/sec)
          r = Mole generation (kmol/sec)
          rx = Mole generation by species (kmol/sec)
          inlet = Inlet stream names or objects
          reserve = Molar holdup name or object
          outlet = Outlet stream name or object
    '''
    name = ''
    V = None
    Q = None
    Qr = None
    r = None
    rx = []
    inlet = ''
    reserve = ''
    outlet = ''
    
class RecoveryObj:
    '''Recovery Object
       The recovery model is a general representation of any separation process.
          split = Split fraction to outlet 1 (0-1)
          inlet = Inlet stream name or objects
          outlet = Outlet stream names or object
    '''
    name = ''
    split = []
    inlet = ''
    outlet = []
    
class SplitterObj:
    '''Splitter Object
       The splitter model keeps the same mole fractions but divides the molar
        flowrate into multiple streams.
          split = Split fraction to outlet 1 (0-1)
          inlet = Inlet stream name or objects
          outlet = Outlet stream names or object
    '''
    name = ''
    split = []
    inlet = ''
    outlet = []    

class StageObj:
    '''Stage (Distillation) Object
       The stage model is one stage of a distillation column
          l_in  = Inlet liquid stream
          l_out = Outlet liquid stream
          v_in  = Inlet vapor stream
          v_out = Outlet vapor stream
          q = Heat addition (+) or loss (-) rate
          dp_in_liq = Pressure drop below stage
          dp_in_vap = Pressure drop above stage
    '''
    name = ''
    l_in = ''
    l_out = ''
    v_in = ''
    v_out = ''
    q = None
    dp_in_liq = None
    dp_in_vap = None
    
class StreamLagObj:
    '''Stream Lag Object
          tau = time constant
          inlet = inlet stream name or object
          outlet = outlet stream name or object
    '''
    name = ''
    tau = None
    inlet = ''
    outlet = ''    

class VesselObj:
    '''Vessel Object
       A vessel that includes inlets, mixing, and an outlet
          V = Volume (m^3)
          Q = Heat input (J/sec)
          inlet = Inlet stream names or objects
          reserve = Molar holdup name or object
          outlet = Outlet stream name or object
    '''
    name = ''
    V = None
    Q = None
    inlet = ''
    reserve = ''
    outlet = ''

class Flowsheet():        
    def __init__(self,m,stream_level=1):
        # record the GEKKO instances where the objects are added
        self.m = m
        self.sl = stream_level # use pressure and energy balance
        if self.sl>=1:
            # use t,p,h,ndot,x[i] with STREAM_LEVEL = 1
            # models such as flash or other VLE require this
            self.m.options.STREAM_LEVEL = 1
        else:
            # use ndot,x[i] with STREAM_LEVEL = 0
            # track compositions only for blending and transport calculations
            self.m.options.STREAM_LEVEL = 0
        return
        
    def add_obj(self,name='',n=None):
        x = name.lower() + str(len(self.m._objects) + 1)
        if n==None:
            self.m._objects.append(x+'='+name)
        else:
            self.m._objects.append(x+'='+name+'('+str(n)+')')
        return x
        
    def dfrac(self):
        '''Generate default fraction based on number of components
        
        dfrac() = 1 / max(1,n_compounds)
        
        The max is used to protect for cases where the number of compounds is zero.
        '''
        return 1.0/max(1.0,float(len(self.m._compounds)))

    def cxn(self,x,val=0.5,cn='',fixed=True):
        '''Check input for expression and create a new variable if not
        a parameter or variable
        
        cxn(x,value=0.5,cn='',fixed=True)
        
        x = the expression or parameter
        val = default value if new parameter or variable is created
        cn = connection name
        fixed = Gekko parameter (True) or variable (False) if None
        '''
        if x==None:
            if fixed:
                x = self.m.Param(val)
            else:
                x = self.m.Var(val)
        else:
            if not isinstance(x,(GKVariable,GKParameter)):
                # create input variable if it is an expression
                xin = self.m.Var(val)
                self.m.Equation(xin==x)
                x = xin
        if cn!='':
            self.m._connections.append(x.name+'='+cn)
        return x

    def cxnl(self,x,val=0.5,cn='',fixed=True):
        '''Check input list for expression and create a new variable if not
        a parameter or variable
        
        cxnl(x,val=0.5,cn='',fixed=True)
        
        x = list of constants, parameters, variables, or expressions
        val = default value if new parameter or variable is created
        cn = connection name base + [1]...[n]
        fixed = Gekko parameter (True) or variable (False) if []
        '''
        nc = len(self.m._compounds)
        if x==[]:
            x = [None]*nc
            for i in range(nc):
                if fixed:
                    x[i] = self.m.Param(val)
                else:
                    x[i] = self.m.Var(val)
        else:
            if len(x)!=nc:
                raise Exception('Error: array length must match number of declared compounds: '\
                                +str(nc))
            for i in range(nc):
                if not isinstance(x[i],(GKVariable,GKParameter)):
                    # create input variable if it is an expression
                    xi = self.m.Var(val)
                    self.m.Equation(xi==x[i])
                    x[i] = xi
        if cn!='':
            for i in range(nc):
               self.m._connections.append(x[i].name+'='+cn+'['+str(i+1)+']')

        return x
        
    def connect(self,s1,s2):
        '''Connect two objects 
        The first name dictates the properties of the combined object.
        
        connect(s1,s2)
        
        s1 = object or name of object 1 (string)
        s2 = object or name of object 2 (string)
        '''
        try:
            c1 = s1.name # if object
        except:
            c1 = s1      # if string
            
        try:
            c2 = s2.name # if object
        except:
            c2 = s2      # if string

        # add connection for streams with * to connect all elements
        self.m._connections.append(c1+'.*='+c2+'.*')

        return
           
    def set_phase(self,y,phase='liquid'):
        '''Set Phase
        set_phase(y,phase='liquid')

        Set phase of a Stream or Reserve Object as 'solid', 'liquid', or 'vapor'
        '''
        if type(y)==str:
           z = y
        else:
           try:
               z = y.name
           except:
               raise Exception('set_phase(y,phase): y must be a ' + \
                               'string or an object with y.name')
        # stream phase
        if phase==None:
            self.m._connections.append(z+'.sfrc=0')
            self.m._connections.append(z+'.lfrc=1')
            self.m._connections.append(z+'.vfrc=0')
            phase = 'liquid'
        else:
            if not (type(phase)==str):
                raise Exception('Phase must be a string: solid, liquid, or vapor')
            if phase.lower()=='solid':
                self.m._connections.append(z+'.sfrc=1')
                self.m._connections.append(z+'.lfrc=0')
                self.m._connections.append(z+'.vfrc=0')
            elif phase.lower()=='liquid':
                self.m._connections.append(z+'.sfrc=0')
                self.m._connections.append(z+'.lfrc=1')
                self.m._connections.append(z+'.vfrc=0')
            elif phase.lower()=='vapor':
                self.m._connections.append(z+'.sfrc=0')
                self.m._connections.append(z+'.lfrc=0')
                self.m._connections.append(z+'.vfrc=1')
            else:
                raise Exception('Phase must be a string: solid, liquid, or vapor')

        return phase

    def reserve(self,fixed=False):
        '''
        reserve(fixed=True)

        Output: Reserve Object
          P = Pressure (Pa)
          T = Temperature (K)
          n = Molar holdup (kmol)
          x = Array of mole fractions
          phase = Phase (solid, liquid, vapor)
          fixed = Gekko parameter (True) or variable (False) if None or []
        '''
        self._thermo_obj = True

        # create stream object
        y = ReserveObj()
        y.name = self.add_obj('Reserve')
        
        if self.sl>=1:
            # pressure
            y.P = self.cxn(y.P,101325.0,y.name+'.P',fixed)
            # temperature
            y.T = self.cxn(y.T,300.0,y.name+'.T',fixed)
            # stream phase
            y.phase = self.set_phase(y,phase=y.phase)
        # molar holdup
        y.n = self.cxn(y.n,1.0,y.name+'.n',fixed)
        # mole fractions
        y.x = self.cxnl(y.x,self.dfrac(),fixed=fixed)
        # additional equation for last mole fraction
        if isinstance(y.x[-1],GKVariable):
            self.m.Equation(y.x[-1]==1-sum(y.x[0:-1]))

        # don't connect last mole fraction to reserve object (explicit calc)
        for i in range(len(self.m._compounds)-1):
            self.m._connections.append(y.x[i].name+'='+y.name+'.x['+str(i+1)+']')
        
        return y   

    def stream(self,fixed=True):
        '''
        y = stream(fixed=True)

        Output: Stream Object 
        y = StreamObj()
          P = Pressure (Pa)
          T = Temperature (K)
          ndot = Molar flow rate (kmol/sec)
          x = Array of mole fractions
          phase = Phase (solid, liquid, vapor)
          fixed = Gekko parameter (True) or variable (False) if None or []
        '''
        self._thermo_obj = True

        # create stream object
        y = StreamObj()
        y.name = self.add_obj('Feed')
        
        if self.sl>=1:
            # pressure
            y.P = self.cxn(y.P,101325.0,y.name+'.P',fixed)
            # temperature
            y.T = self.cxn(y.T,300.0,y.name+'.T',fixed)
            # stream phase
            y.phase = self.set_phase(y,phase=y.phase)
        # molar flow
        y.ndot = self.cxn(y.ndot,1.0,y.name+'.ndot',fixed)
        # mole fractions
        y.x = self.cxnl(y.x,self.dfrac(),fixed=fixed)
        # additional equation for last mole fraction
        if isinstance(y.x[-1],GKVariable):
            self.m.Equation(y.x[-1]==1-sum(y.x[0:-1]))

        # don't connect last mole fraction to stream object (explicit calc)
        for i in range(len(self.m._compounds)-1):
            self.m._connections.append(y.x[i].name+'='+y.name+'.x['+str(i+1)+']')
        
        return y

    def flash(self):
        '''
        y = flash()

        Output: Flash object
          P = Pressure (Pa)
          T = Temperature (K)
          Q = Heat input (J/sec)
          gamma = Activity coefficients for each compound
          inlet = inlet stream name
          vapor = vapor outlet stream name
          liquid = liquid outlet stream name
        '''
        self._thermo_obj = True
            
        # create flash object
        y = FlashObj()
        y.name = self.add_obj('Flash')
        
        if self.sl==0:
            raise Exception('Stream level required >=1 for flash calculation')
        
        # names of inlet stream (1) and outlet streams (2)
        y.inlet = y.name + '.inlet'
        y.vapor = y.name + '.outlet_vap'
        y.liquid = y.name + '.outlet_liq'

        # pressure
        y.P = self.cxn(y.P,101325.0,y.inlet+'.P')
        # temperature
        y.T = self.cxn(y.T,300.0,y.liquid+'.T',fixed=False)        
        # heat input
        y.Q = self.cxn(y.Q,0.0,y.name+'.Q')
        # activity coefficients
        y.gamma = self.cxnl(y.gamma,1.0,y.name+'.gamma')
        
        return y

    def flash_column(self):
        '''
        y = flash_column()

        Output: Flash object
          P = Pressure (Pa)
          T = Temperature (K)
          Q = Heat input (J/sec)
          n = Holdup (kmol)
          gamma = Activity coefficients for each compound
          inlet = inlet stream name
          vapor = vapor outlet stream name
          liquid = liquid outlet stream name
        '''
        self._thermo_obj = True
            
        # create flash_column object
        y = FlashObj()
        y.name = self.add_obj('Flash_Column')
       
        if self.sl==0:
            raise Exception('Stream level required >=1 for flash calculation')

        # names of inlet stream (1) and outlet streams (2)
        y.inlet = y.name + '.feed'
        y.reserve = y.name + '.holdup.reserve'
        y.vapor = y.name + '.flash.outlet_vap'
        y.liquid = y.name + '.flash.outlet_liq'
        
        # pressure
        y.P = self.cxn(y.P,101325.0,y.inlet+'.P')
        # temperature
        y.T = self.cxn(y.T,300.0,y.liquid+'.T',fixed=False)        
        # heat input
        y.Q = self.cxn(y.Q,0.0,y.name+'.flash.Q')
        # molar holdup
        y.n = self.cxn(y.n,1.0,y.reserve+'.n')
        # activity coefficients
        y.gamma = self.cxnl(y.gamma,1.0,y.name+'.flash.gamma')
        
        return y
    
    def mass(self,y=None,rn=''):
        '''
        mass(mo=None)

        Inputs:
          y = Mass Object (mo)
            m = mass (kg)
            mx = mass of components (kg)
            reserve = ''
          rn = Reserve name if already created
        '''
        self._thermo_obj = True
        if y==None:
            y = MassObj()
            
        # create mass object
        y.name = self.add_obj('Mass')
                
        # reserve
        if rn=='':
            y.reserve = self.reserve(fixed=True)
            self.connect(y.reserve,y.name+'.acc')            
        else:
            # connection
            self.connect(rn,y.name+'.acc')
            y.reserve = rn

        # mass
        y.m = self.cxn(y.m,1.0,y.name + '.m',fixed=False)
        
        return y
        
    def massflow(self,y=None,sn=''):
        '''
        massflow(mo=None)

        Inputs:
          y = Massflow Object
            mdot = massflow (kg)
            stream = ''
          sn = Stream name if already created
        '''
        self._thermo_obj = True
        if y==None:
            y = MassflowObj()
            
        # create mass object
        y.name = self.add_obj('Massflow')
                
        # stream
        if sn=='':
            y.stream = self.stream(fixed=True)
            self.connect(y.stream,y.name+'.stream')
        else:
            # connection
            self.connect(sn,y.name+'.stream')
            y.stream = sn

        # massflow
        y.mdot = self.cxn(y.mdot,1.0,y.name + '.mdot',fixed=False)
        
        return y
        
    def massflows(self,y=None,sn=''):
        '''
        massflows(mo=None)

        Inputs:
          y = Massflow Object
            mdot = massflow (kg)
            mdoti = massflow of components (kg)
            stream = ''
          sn = Stream name if already created
        '''
        self._thermo_obj = True
        if y==None:
            y = MassflowObj()
            
        # create mass object
        y.name = self.add_obj('Massflows')
                
        # stream
        if sn=='':
            y.stream = self.stream(fixed=True)
            self.connect(y.stream,y.name+'.stream')
        else:
            # connection
            self.connect(sn,y.name+'.stream')
            y.stream = sn

        # massflow
        y.mdot = self.cxn(y.mdot,1.0,y.name + '.mdot',fixed=False)
        # component massflow
        y.mdoti = self.cxnl(y.mdoti,self.dfrac(),y.name + '.mdoti',fixed=False)
        
        return y

    def molarflows(self,y=None,sn=''):
        '''
        molarflows(mo=None)

        Inputs:
          y = Molarflow Object
            ndot = molarflow (kmol)
            ndoti = molarflow of components (kmol)
            stream = ''
          sn = Stream name if already created
        '''
        self._thermo_obj = True
        if y==None:
            y = MolarflowObj()
            
        # create mass object
        y.name = self.add_obj('Molarflows')
                
        # stream
        if sn=='':
            y.stream = self.stream(fixed=True)
            self.connect(y.stream,y.name+'.stream')
        else:
            # connection
            self.connect(sn,y.name+'.stream')
            y.stream = sn

        # component molarflows
        y.ndoti = self.cxnl(y.ndoti,self.dfrac(),y.name + '.ndoti',fixed=False)
        
        return y   

    def mixer(self,ni=2):
        '''
        mixer(ni=2)

        Inputs:
          ni = Number of inlets (default=2)
        
        Output:
          y = Mixer object
            inlet = inlet stream names
            outlet = outlet stream name
        '''
        self._thermo_obj = True

        # create object
        y = MixerObj()
        y.name = self.add_obj('Mixer',ni)
            
        # names of streams
        y.inlet = [None]*ni
        for i in range(ni):
           y.inlet[i] = y.name+'.inlet['+str(i+1)+']'
        y.outlet = y.name+'.outlet'
        
        return y
        
        
    def pid(self):
        '''PID (Proportional Integral Derivative) Object
              co = Controller Output (u)
              pv = Process Variable (y)
              sp = Process Variable Set Point (ysp)
              Kc = PID Proportional constant
              tauI = PID Integral constant
              tauD = PID Derivative constant
              i = Integral term  
           
           Description:  PID: Proportional Integral Derivative Controller
           In the frequency domain the PID controller is described by
            U(s) = Kc*Y(s) + Y(s)*Kc/s*tauI + Kc*taud*s*Y(s)
           In the time domain the PID controller is described by
            u(t) = Kc*(ysp-y(t)) + (Kc/taui)*Integral(t=0...t)(ysp-y(t))dt + Kc*taud*dy(t)/dt
           This implementation splits the single equation into two equations
           The second equation is necessary to avoid the numerical integration.
           The equations are posed in an open equation form.  The integral time
            constant is multiplied through to avoid potential divide by zero.
           This form may have an advantage over placing the term taui in with the
            integral equation for cases where taui becomes very small.
            0 = -u*taui + Kc*((ysp-y)*taui + Integral + taud*(dy/dt)*taui)
            0 = d(Integral)/dt - (ysp-y)          
        '''
        self._thermo_obj = True
            
        # create PID object
        y = PIDObj()
        y.name = self.add_obj('PID')
                
        # connect parameters and variables
        y.co = self.cxn(y.co,0.0,y.name + '.u',fixed=False)
        y.pv = self.cxn(y.pv,0.0,y.name + '.y',fixed=False)
        y.sp = self.cxn(y.sp,0.0,y.name + '.ysp',fixed=True)
        y.Kc = self.cxn(y.Kc,0.0,y.name + '.Kc',fixed=True)
        y.tauI = self.cxn(y.tauI,0.0,y.name + '.tauI',fixed=True)
        y.tauD = self.cxn(y.tauD,0.0,y.name + '.tauD',fixed=True)
        y.i = self.cxn(y.i,0.0,y.name + '.i',fixed=False)
        
        return y
        
    def pump(self):
        '''Pump Object
              dp = change in pressure
              inlet = inlet stream name or object
              outlet = outlet stream name or object
        '''
        self._thermo_obj = True
            
        # create object
        y = PumpObj()
        y.name = self.add_obj('Pump')
            
        y.dp = self.cxn(y.dp,0.0,y.name + '.dp',fixed=True)
        
        # names of streams
        y.inlet = y.name+'.inlet'
        y.outlet = y.name+'.outlet'
        
        return y
                        
    def reactor(self,ni=1):
        '''
        reactor(ni=1)
           
        Inputs:
          ni = Number of inlet streams (default=1)
        
        Output:
          y = Reactor object
            V = Volume (m^3)
            Q = Heat input (J/sec)
            Qr = Heat generation by reaction (J/sec)
            r = Mole generation (kmol/sec)
            rx = Mole generation by species (kmol/sec)
            inlet = Inlet stream names or objects
            reserve = Molar holdup name or object
            outlet = Outlet stream name or object

           Reactor Object
              The reactor model is equivalent to the vessel model but also
              includes generation of species through reaction.  The reaction rates
              are defined as (+) generation and (-) consumption.
           
           In addition to the reaction rates, there is a term for heat generation
              from exothermic reactions (+) or heat removal from endothermic reactions (-)
        '''
        self._thermo_obj = True
            
        # create object
        y = ReactorObj()
        y.name = self.add_obj('Reactor',ni)

        # V = Volume (m^3)
        y.V = self.cxn(y.V,1.0,y.name+'.V',fixed=True)
        # Q = Heat input (J/sec)
        y.Q = self.cxn(y.Q,0.0,y.name+'.Q',fixed=True)
        # Qr = Heat generation by reaction (J/sec)
        y.Qr = self.cxn(y.Qr,0.0,y.name+'.Qr',fixed=True)
        # r = Mole generation (kmol/sec)
        y.r = self.cxn(y.r,0.0,y.name+'.r',fixed=False)
        # rx = Mole generation by species (kmol/sec)
        y.rx = self.cxnl(y.rx,0.0,y.name+'.r',fixed=True)

        # names of streams and reserve
        y.inlet = [None]*ni
        for i in range(ni):
           y.inlet[i] = y.name+'.inlet['+str(i+1)+']'
        y.reserve = y.name+'.acc'
        y.outlet = y.name+'.outlet'
        
        return y

    def recovery(self):
        '''
        recovery()       
                
        Output:
          y = Recovery Object
             split = Split fraction to outlet 1 (0-1)
             inlet = Inlet stream name or objects
             outlet = Outlet stream names or object

        Recovery Object
          The recovery model is a general representation of any separation process.
        '''
        self._thermo_obj = True
            
        # create object
        y = RecoveryObj()
        y.name = self.add_obj('Recovery')

        # split = Split fraction to outlet 1 (0-1)
        y.split = self.cxnl(y.split,1.0,y.name+'.split',fixed=True)

        # names of streams and reserve
        y.inlet = y.name+'.inlet'
        y.outlet = [y.name+'.outlet['+str(i+1)+']' for i in range(2)]
        
        return y

    def splitter(self,no=2):
        '''
        splitter()

        Input:
          no = Number of outlet streams 
                
        Output:
          y = Splitter Object
             split = Split fraction to each outlet
             inlet = Inlet stream name or objects
             outlet = Outlet stream names or object

        Splitter Object
        The splitter model keeps the same mole fractions but divides the molar
         flowrate into multiple streams.
        '''
        self._thermo_obj = True
            
        # create object
        y = SplitterObj()
        no = max(2,no)
        y.name = self.add_obj('Splitter',no)

        # split = Split fraction to outlets (0-1)
        y.split = [None]*no
        sf = 1.0/no
        for i in range(no):
            if i!=no-1:
                y.split[i] = self.m.Param(sf)
            else:
                # last split fraction is a variable
                y.split[i] = self.m.Var(sf)
            cn = y.name+'.split['+str(i+1)+']'
            self.m._connections.append(y.split[i].name+'='+cn)

        # names of streams and reserve
        y.inlet = y.name+'.inlet'
        y.outlet = [y.name+'.outlet['+str(i+1)+']' for i in range(no)]
        
        return y
        
    def stage(self,opt=2):
        '''
        stage(opt=2)

        Input:
          opt = Option for stage type (1=Index-1 DAE, 2=Index-2 DAE) 
                
        Output:
          y = Stage Object
             l_in  = Inlet liquid stream
             l_out = Outlet liquid stream
             v_in  = Inlet vapor stream
             v_out = Outlet vapor stream
             q = Heat addition (+) or loss (-) rate
             dp_in_liq = Pressure drop below stage
             dp_in_vap = Pressure drop above stage
        Stage Object
        The stage model is one stage (tray, packing height) of a distillation column.
        '''
        self._thermo_obj = True
            
        # create object
        y = StageObj()
        if opt==2:
           y.name = self.add_obj('stage_2')
        else:
           y.name = self.add_obj('stage_1')        

        # names of streams
        y.l_in = y.name+'.l_in'
        y.l_out = y.name+'.l_out'
        y.v_in = y.name+'.v_in'
        y.v_out = y.name+'.v_out'
        
        if opt==2:
            y.q = self.cxn(y.q,0.0,y.name + '.q',fixed=True)
            y.dp_in_liq = self.cxn(y.dp_in_liq,0.0,y.name + '.dp_in_liq',fixed=True)
            y.dp_in_vap = self.cxn(y.dp_in_vap,0.0,y.name + '.dp_in_vap',fixed=True)
        
        return y

    def stream_lag(self):
        '''
        stream_lag()
        
        Stream Lag Object
          Output
              tau = time constant (sec)
              inlet = inlet stream name or object
              outlet = outlet stream name or object
        '''
        self._thermo_obj = True
            
        # create object
        y = StreamLagObj()
        y.name = self.add_obj('Stream_Lag')
            
        y.tau = self.cxn(y.tau,1.0,y.name + '.tau',fixed=True)
        
        # names of streams
        y.inlet = y.name+'.inlet'
        y.outlet = y.name+'.outlet'
        
        return y

    def vessel(self,ni=1,mass=False):
        '''
        reactor(ni=1,mass=False)
           
        Inputs:
          ni = Number of inlets (default=1)
          mass = Mass flows instead of molar flows (default=False)
        
        Output:
          y = Reactor object
            V = Volume (m^3)
            Q = Heat input (J/sec)
            inlet = Inlet stream names or objects
            reserve = Molar holdup name or object
            outlet = Outlet stream name or object

           Reactor Object
              The reactor model is equivalent to the vessel model but also
              includes generation of species through reaction.  The reaction rates
              are defined as (+) generation and (-) consumption.
           
           In addition to the reaction rates, there is a term for heat generation
              from exothermic reactions (+) or heat removal from endothermic reactions (-)
        '''
        self._thermo_obj = True
            
        # create object
        y = ReactorObj()
        if mass:
            y.name = self.add_obj('Vesselm',ni)
        else:
            y.name = self.add_obj('Vessel',ni)

        # V = Volume (m^3)
        y.V = self.cxn(y.V,1.0,y.name+'.V',fixed=True)
        # Q = Heat input (J/sec)
        y.Q = self.cxn(y.Q,0.0,y.name+'.Q',fixed=True)

        # names of streams and reserve
        y.inlet = [None]*ni
        for i in range(ni):
           y.inlet[i] = y.name+'.inlet['+str(i+1)+']'
        y.reserve = y.name+'.acc'
        y.outlet = y.name+'.outlet'
        
        return y




"""
global_option_list = ['APPINFO', 'APPINFOCHG', 'APPSTATUS', 'BNDS_CHK', 'COLDSTART',
                      'CSV_READ', 'CSV_WRITE', 'CTRLMODE', 'CTRL_HOR', 'CTRL_TIME',
                      'CTRL_UNITS', 'CV_TYPE', 'CV_WGT_SLOPE', 'CV_WGT_START',
                      'CYCLECOUNT', 'DBS_READ', 'DBS_WRITE', 'DIAGLEVEL', 'EV_TYPE',
                      'EV_WGT_SLOPE', 'HIST_HOR', 'HIST_UNITS', 'ICD_CALC', 'IMODE',
                      'ITERATIONS', 'MAX_ITER', 'MAX_TIME', 'MEAS_CHK', 'MV_DCOST_SLOPE',
                      'MV_STEP_HOR', 'MV_TYPE', 'NODES', 'OBJFCNVAL', 'OTOL', 'PRED_HOR',
                      'PRED_TIME', 'REDUCE', 'REQCTRLMODE', 'RTOL', 'SCALING', 'SENSITIVITY',
                      'SOLVESTATUS', 'SOLVER', 'SOLVETIME', 'SPECS', 'TIME_SHIFT',
                      'WEB', 'WEB_MENU', 'WEB_REFRESH']
"""                      
global_options_inputs =  ['AUTO_COLD', 'BNDS_CHK', 'CSV_READ', 'CSV_WRITE', 
'CTRL_UNITS', 'CV_WGT_SLOPE', 'CV_WGT_START', 'CV_TYPE', 'DBS_LEVEL', 
'DBS_READ', 'DBS_WRITE', 'DIAGLEVEL', 'EV_WGT_SLOPE', 'EV_TYPE', 'FILTER', 
'FRZE_CHK', 'HIST_HOR', 'HIST_UNITS', 'ICD_CALC', 'IMODE', 'LINEAR', 'MAX_ITER', 
'MAX_MEMORY', 'MAX_TIME', 'MEAS_CHK', 'MV_DCOST_SLOPE', 'MV_STEP_HOR', 
'MV_TYPE', 'NODES', 'OTOL', 'REDUCE', 'REPLAY', 'REQCTRLMODE', 'RTOL', 
'SCALING', 'SENSITIVITY', 'SEQUENTIAL', 'SOLVER', 'SPECS', 'SPC_CHART', 
'STREAM_LEVEL', 'TIME_SHIFT', 'WEB', 'WEB_MENU', 'WEB_REFRESH', 'WEB_PLOT_FREQ']

global_options_outputs= ['APPINFO', 'APPINFOCHG', 'APPSTATUS', 'CTRLMODE', 
'ITERATIONS', 'OBJFCNVAL', 'SOLVESTATUS', 'SOLVETIME']

global_options_inout = ['BAD_CYCLES', 'COLDSTART', 'CTRL_HOR', 'CTRL_TIME', 
'CYCLECOUNT', 'PRED_HOR', 'PRED_TIME']



class GKGlobalOptions():
    """Holds the global options for the application"""

    def __init__(self):
        # prevents the __setattr__ function from sending options to the server
        # until the __init__ function has completed since they should only be
        # sent if changed from their defaults
        self.__dict__['_initialized'] = False
        
        self._input_option_list = global_options_inputs
        self._output_option_list = global_options_outputs
        self._inout_option_list = global_options_inout

        # set defaults for all global options
        #inputs 
        self.AUTO_COLD = 0
        self.BNDS_CHK = 1
        self.CSV_READ = 2
        self.CSV_WRITE = 1
        self.CTRL_UNITS = 1
        self.CV_TYPE = 1
        self.CV_WGT_SLOPE = 0.0
        self.CV_WGT_START = 0
        self.DBS_LEVEL = 1
        self.DBS_READ = 1
        self.DBS_WRITE = 1
        self.DIAGLEVEL = 0
        self.EV_TYPE = 1
        self.EV_WGT_SLOPE = 0.0
        self.FILTER = 1
        self.FRZE_CHK = 1
        self.HIST_HOR = 0
        self.HIST_UNITS = 0
        self.ICD_CALC = 0
        self.IMODE = 3
        self.LINEAR = 0
        self.MAX_ITER = 250
        self.MAX_MEMORY = 4
        self.MAX_TIME = 1.0e20
        self.MEAS_CHK = 1
        self.MV_DCOST_SLOPE = 0.1
        self.MV_STEP_HOR = 1
        self.MV_TYPE = 0
        self.NODES = 2
        self.OTOL = 1.0e-6
        self.REDUCE = 0
        self.REQCTRLMODE = 3
        self.REPLAY = 0
        self.RTOL = 1.0e-6
        self.SCALING = 1
        self.SENSITIVITY = 0
        self.SEQUENTIAL = 0
        self.SOLVER = 3
        self.SPC_CHART = 0
        self.SPECS = 1
        self.STREAM_LEVEL = 0
        self.TIME_SHIFT = 1
        self.WEB = 0
        self.WEB_MENU = 1
        self.WEB_PLOT_FREQ = 1
        self.WEB_REFRESH = 10
        #outputs
        self.APPINFO = 0
        self.APPINFOCHG = 0
        self.APPSTATUS = 1
        self.CTRLMODE = 1
        self.ITERATIONS = 1
        self.OBJFCNVAL = 0.0
        self.SOLVESTATUS = 1
        self.SOLVETIME = 1.0
        #inputs/outputs
        self.BAD_CYCLES = 0
        self.COLDSTART = 0
        self.CTRL_HOR = 1
        self.CTRL_TIME = 60
        self.CYCLECOUNT = 0
        self.PRED_HOR = 1.0
        self.PRED_TIME = 60.0
        

        # now allow options to be sent to the server
        self._initialized = True

    def getOverridesString(self):
        ''' Returns string to go in dbs file

            Example return value:
                NLC.APPINFO = 0
                APPINFOCHG  = 0
                ...
                NLC.WEB_REFRESH = 10
        '''
        result = ''

        for attr, value in self.__dict__.items():
            # If the attribute is in the list of exceptions, do not print
            if(attr in global_options_inputs+global_options_inout):
                result = result + "APM." + attr + " = " + str(value) + "\n"

        return result



    def __str__(self):
        for attr, value in self.__dict__.iteritems():
            print(str(attr) + ' ' + str(value))
        return "This is the string function"

    def __setattr__(self, name, value):
        if self._initialized:
            #ignore cases on global options
            name = name.upper()

            #only allow user to set input or input/output options:
            if name in global_options_inputs+global_options_inout:
                self.__dict__[name] = value
                    
            #don't allow writing to output properties by default
            elif name in global_options_outputs:
                #define outputs by passing list/tuple with 1st element being True
                #to override the output writing prevention 
                try:
                    if value[0] == True:
                        self.__dict__[name] = value[1]
                    else:
                        raise TypeError
                except TypeError:
                    print(str(name)+" is an output property")
                    raise AttributeError
                    
            #no other properties allowed
            else:
                print(str(name)+" is not a recognized property")
                raise AttributeError
                
        #for initializing model
        else:
            self.__dict__[name] = value

    #make attributes case in-sensitive for reading too
    def __getattr__(self,name):
        name_upp = name.upper()
        try:
            return self.__dict__[name_upp]
        except:
            raise AttributeError(str(name)+ ' is not a valid model option')


import __main__ as main

import json
import logging
import os
import socket
import sys
import threading
import webbrowser


from flask import Flask, jsonify, redirect

from flask_cors import CORS

from .gk_operators import GK_Intermediate
from .gk_parameter import GKParameter, GK_MV, GK_FV
from .gk_variable import GKVariable, GK_CV, GK_SV

# Toggle development and production modes
DEV = False
WATCHDOG_TIME_LENGTH = 0

if DEV:
    # It will leave hanging processes if it gets killed in dev mode
    WATCHDOG_TIME_LENGTH = 10000000
    # This command can be used in the event that there is a
    # hanging process: sudo fuser -n tcp -k 8050
else:
    WATCHDOG_TIME_LENGTH = 4
    log = logging.getLogger('werkzeug')
    # This little nugget sets flask to only log http errors, very important
    # when we are polling every second.
    log.setLevel(logging.ERROR)

# Initialize the flask api
app = Flask(__name__, static_url_path='/gui/dist')

# Disable CORS as we are just communicating with localhost on different ports
CORS(app)


def watchdog_timer():
    """
    Exit the process. Used if watchdog timer expires.

    This acts as a watchdog timer. If the front-end does not make a call at
    least every timeout seconds then the main process is stopped. Every time
    this method is called the timer gets reset.
    Could normally use signal.alarm, but need to support windows
    """
    print('Browser display closed. Exiting...')
    print('If not running in an interactive shell, exit with ^C.')
    # os._exit(0)
    sys.exit()


class FlaskThread(threading.Thread):
    """
    Flask API thread. Pulls the required data from options.json and
    results.json and displays by opening the local browser to the Vue app.
    """
    def __init__(self, path, debug, port):
        threading.Thread.__init__(self)
        self.history_horizon = 5     # History horizon that will be displayed on the plot
        self.has_data = False        # Variable defines if the Gekko data is loaded
        self.has_new_update = False
        self.path = path             # Path to tmp dir where Gekko output is
        self.debug = debug
        self.port = port

        self.gekko_data = {}         # Combined, final Gekko data object
        self.options = {}            # Dict loaded from options.json
        self.results = {}            # Dict loaded from results.json

        self.vars_dict = {}          # vars dict with script names as keys
        self.make_vars_map()         # sets map of script names to apmonitor names
        self.options_dict = {}       # options dict with script names as keys
        self.model = {}              # APM model information
        self.info = {}
        self.get_script_data()
        self.has_data = True

        # This is used for tvars_maphe polling between the api and the Vue app
        self.alarm = threading.Timer(WATCHDOG_TIME_LENGTH, watchdog_timer)

    # Parameters require a little special handling, only called from get_var_from_main
    def get_parameter_from_main(self, param):
        """Special handling for GK_Parameters"""
        main_dict = vars(main)
        data = list(filter(lambda d: d['name'] == param, self.gekko_data['vars']['parameters']))[0]
        try:
            data['data'] = self.results[main_dict[param].name]
            data['x'] = self.results['time']
            data['options'] = self.options[main_dict[param].name]
        except KeyError:
            # Some vars are not in options.json and so do not make it into self.options
            # This case should be safe to ignore
            pass
        except Exception:
            print("Error getting data for: %s (%s)" % (param, main_dict[param].name))

        ## historical data
        if isinstance(main_dict[param], (GK_MV, GK_FV)):
            data_hist = list(filter(lambda d: d['name'] == param + '_hist', self.gekko_data['vars']['parameters']))[0]
            data_hist['data'] = data_hist['data'] + [self.results[main_dict[param].name][0]]
            timestep = self.results['time'][1] - self.results['time'][0]
            data_hist['x'] = [data_hist['x'][0] - timestep] + data_hist['x']
                
    def get_variable_fron_main(self, variable):
        """Special handling for GK_Variables"""
        main_dict = vars(main)
            
        ## Store historical data
        timestep = self.results['time'][1] - self.results['time'][0]
        
        if isinstance(main_dict[variable], (GK_CV, GK_SV)):
            if main_dict[variable].name + '.bcv' in self.results:
                #biased history
                var_hist_bias = list(filter(
                    lambda d: d['name'] == variable + '_hist(bias)', self.gekko_data['vars']['variables']))[0]
                var_hist_bias['data'] = var_hist_bias['data'] + [self.results[main_dict[variable].name + '.bcv'][0]] #store data value
                var_hist_bias['x'] = [var_hist_bias['x'][0] - timestep] + var_hist_bias['x']    #update time array
                #unbiased history
                var_hist_nobias = list(filter(
                    lambda d: d['name'] == variable + '_hist(nobias)', self.gekko_data['vars']['variables']))[0]
                var_hist_nobias['data'] = var_hist_nobias['data'] + [self.results[main_dict[variable].name][0]] #store data value
                var_hist_nobias['x'] = [var_hist_nobias['x'][0] - timestep] + var_hist_nobias['x']    #update time array
            
        
        ## Plot prediction from current solve
        var = list(filter(lambda d: d['name'] == variable, self.gekko_data['vars']['variables']))[0]
        try:
            if isinstance(main_dict[variable], (GK_CV, GK_SV)):
                if main_dict[variable].name + '.bcv' in self.results:
                    var['data'] = self.results[main_dict[variable].name + '.bcv']
                    var['x'] = self.results['time']
                    var['options'] = self.options[main_dict[variable].name]
                    
                    var_nobias = list(filter(
                    lambda d: d['name'] == variable + '(nobias)', self.gekko_data['vars']['variables']))[0]
                    var_nobias['data'] = self.results[main_dict[variable].name]
                    var_nobias['x'] = self.results['time']
                
                else:
                    var['data'] = self.results[main_dict[variable].name]
                    var['x'] = self.results['time']
                    var['options'] = self.options[main_dict[variable].name]
            else:
                var['data'] = self.results[main_dict[variable].name]
                var['x'] = self.results['time']
                var['options'] = self.options[main_dict[variable].name]
        except KeyError:
            # Some vars are not in options.json and so do not make it into self.options
            # This case should be safe to ignore
            pass
        except Exception:
            print("Error getting data for: %s (%s)" % (variable, main_dict[variable].name))
            
            
        if main_dict[variable].name + '.tr_hi' in self.results:
            data = list(filter(lambda d: d['name'] == variable + '(Tr_hi)', self.gekko_data['vars']['variables']))[0]
            # Replace the list entirely as we want a new trajectory for each solve
            data['data'] = self.results[main_dict[variable].name + '.tr_hi']
            data['x'] = self.results['time']
        if main_dict[variable].name + '.tr_lo' in self.results:
            data = list(filter(lambda d: d['name'] == variable + '(Tr_lo)', self.gekko_data['vars']['variables']))[0]
            # Replace the list entirely as we want a new trajectory for each solve
            data['data'] = self.results[main_dict[variable].name + '.tr_lo']
            data['x'] = self.results['time']
            
        if main_dict[variable].name + '.tr' in self.results:
            data = list(filter(lambda d: d['name'] == variable + '(Tr)', self.gekko_data['vars']['variables']))[0]
            # Replace the list entirely as we want a new trajectory for each solve
            data['data'] = self.results[main_dict[variable].name + '.tr']
            data['x'] = self.results['time']



    def get_var_from_main(self, var):
        """Gets data about a variable and packs it into gekko_data"""
        main_dict = vars(main)

        # Set the variable dict if this is the first time
        if not self.has_data:
            options = {}
            try:
                options = self.options[main_dict[var].name]
            except:
                options = {}
            try:
                var_dict = {
                    'name': var,
                    'data': [self.results[main_dict[var].name][0]],
                    'x': [self.results['time'][0]],
                    'options': options
                }
            except Exception:
                var_dict = {
                    'name': var,
                    'data': [],
                    'x': [],
                    'options': options
                }
            if isinstance(main_dict[var], GKVariable):
                self.gekko_data['vars']['variables'].append(var_dict)
                if main_dict[var].name + '.tr_hi' in self.results:
                    d = var_dict.copy() 
                    d['name'] = var + '(Tr_hi)'
                    self.gekko_data['vars']['variables'].append(d)
                if main_dict[var].name + '.tr_lo' in self.results:
                    d = var_dict.copy()
                    d['name'] = var + '(Tr_lo)'
                    self.gekko_data['vars']['variables'].append(d) 
                if main_dict[var].name + '.tr' in self.results:
                    d = var_dict.copy()
                    d['name'] = var + '(Tr)'
                    self.gekko_data['vars']['variables'].append(d) 
                if main_dict[var].name + '.bcv' in self.results:
                    d1 = var_dict.copy()
                    d1['name'] = var + '_hist(bias)'
                    self.gekko_data['vars']['variables'].append(d1)
                    d2 = var_dict.copy()
                    d2['name'] = var + '_hist(nobias)'
                    self.gekko_data['vars']['variables'].append(d2)
                    d3 = var_dict.copy()
                    d3['name'] = var + '(nobias)'
                    self.gekko_data['vars']['variables'].append(d3)
            elif isinstance(main_dict[var], GKParameter):
                self.gekko_data['vars']['parameters'].append(var_dict)
                if isinstance(main_dict[var], (GK_MV, GK_FV)):
                    d = var_dict.copy()
                    d['name'] = var + '_hist'
                    self.gekko_data['vars']['parameters'].append(d)
            elif isinstance(main_dict[var], GK_Intermediate):
                self.gekko_data['vars']['intermediates'].append(var_dict)

        # Update the variable if the data has been loaded before
        data = False
        if isinstance(main_dict[var], GKVariable):
            self.get_variable_fron_main(var)
            return
        elif isinstance(main_dict[var], GKParameter):
            self.get_parameter_from_main(var)
            return
        elif isinstance(main_dict[var], GK_Intermediate):
            data = list(filter(lambda d: d['name'] == var, self.gekko_data['vars']['intermediates']))[0]
        try:
            data['data'] = data['data'] + [self.results[main_dict[var].name][0]]
            data['options'] = self.options[main_dict[var].name]

        except KeyError:
            # Some vars are not in options.json and so do not make it into self.options
            # This case should be safe to ignore
            pass
        except Exception as e:
            raise e
                
    def make_vars_map(self):
        """Maps python script names to APMonitor names"""
        vars_map = {}
        main_dict = vars(main)
        for var in main_dict:
            if isinstance(main_dict[var], (GKVariable,GKParameter,GK_Intermediate)):
                vars_map[main_dict[var].name] = var
            if isinstance(main_dict[var], list):
                list_var = main_dict[var]
                for i in range(len(list_var)):
                    if isinstance(list_var[i], (GKVariable,GKParameter,GK_Intermediate)):
                        vars_map[list_var[i].name] = var+'['+str(i)+']'
        self.vars_map = vars_map

    def get_options(self):
        """Generates options_dict from results and options"""
        for var in self.vars_map:
            if var != 'time':
                if DEV:
                    print('Mapped', var, 'to', self.vars_map[var])
                try:
                    self.vars_dict[self.vars_map[var]] = self.results[var]
                except:
                    # FIXME: Check to make sure passing this error will not cause unintended errors
                    pass
                for v in self.vars_map:
                    try:
                        self.options_dict[self.vars_map[v]] = self.options[v]
                    except:
                        # Some vars are not in options.json, but only have values in results.json
                        pass

    def get_script_data(self):
        """Gather the data that GEKKO returns from the run and process it into
        the objects that the GUI can handle. Only run on initialization"""
        # When calling `m.solve()` in a loop the Gui will be initialized before
        # `m.solve()` is ever called, so the files might not be there.
        try:
            # Load options.json
            self.options = json.loads(open(os.path.join(self.path,'options.json')).read())
            # Load results.json
            self.results = json.loads(open(os.path.join(self.path,"results.json")).read())
            self.gekko_data['model'] = self.options['APM']
            self.gekko_data['info'] = self.options['INFO']
            self.gekko_data['time'] = self.results['time']
            self.gekko_data['vars'] = {}
            self.gekko_data['vars']['variables'] = []
            self.gekko_data['vars']['parameters'] = []
            self.gekko_data['vars']['constants'] = []
            self.gekko_data['vars']['intermediates'] = []

            self.vars_dict['time'] = self.results['time']
            self.model = self.options['APM']

            main_dict = vars(main)
            for var in main_dict:
                if (var != 'time') and isinstance(main_dict[var], (GKVariable, GKParameter, GK_Intermediate)):
                    self.get_var_from_main(var)
            self.get_options()

            self.has_model_data = True

        except Exception as e:
            raise e

    def handle_api_call(self, data):
        """Handles the generic aspects of all incoming API calls"""
        try:
            # When solved in a loop there might not be any data to send.
            if self.has_model_data:
                resp = jsonify(data)
                resp.headers.add('Access-Control-Allow-Origin', '*')
                # resp.headers.add('Access-Control-Allow-Origin', 'http://localhost:8080')
            else:
                # This will let the Vue app know that no data has arrived yet
                resp = jsonify({'dataAvailable': False})
                resp.headers.add('Access-Control-Allow-Origin', '*')
        except Exception as e:
            app.unhandled_error(e)

        # This resets the watchdog timer, kind of a hack, but it works
        self.alarm.cancel()
        self.alarm = threading.Timer(WATCHDOG_TIME_LENGTH, watchdog_timer)
        self.alarm.start()
        return resp

    def set_endpoints(self):
        """Sets the flask API endpoints"""
        try:
            # pylint: disable=W0612
            @app.route('/data')
            def get_data():
                self.has_new_update = False
                return self.handle_api_call(self.gekko_data)

            @app.route('/get_options')
            def get_options():
                return self.handle_api_call(self.options_dict)

            @app.route('/poll')
            def get_poll():
                return self.handle_api_call({'updates': self.has_new_update})

            @app.route('/<path:path>')
            def static_file(path):
                return app.send_static_file(path)

            @app.route('/')
            def redirect_to_index():
                newPath = 'http://localhost:' + str(self.port) + '/index.html'
                return redirect(newPath)

            @app.errorhandler(404)
            def endpoint_not_found(error):
                return jsonify(error=404, text='Endpoint not found')

            @app.errorhandler(Exception)
            def unhandled_error(e):
                return jsonify(error=500, text='Internal API Error')

        except AssertionError as e:
            # This try/except is because ipython keeps the endpoints in
            # memory between runs. This will simply not set the endpoints if they
            # are already there.
            pass

        except Exception as e:
            raise e

    def run(self):
        """Starts up the Flask API

        Called by FlaskThread.start() as that is how python threads work
        """
        self.set_endpoints()
        # Debug in flask does not work when run on a separate thread
        app.run(debug=False, port=self.port, threaded=True)
        self.alarm.start()

    def check_history_horizon(self):
        """Checks if a Variable is overflowing the history_horizon and trims
        it as necessary if it is."""
        for var_list in self.gekko_data['vars']:
            for var in var_list:
                if len(var) >= self.history_horizon:
                    self.gekko_data['vars'][var_list][var] = self.gekko_data['vars'][var_list][var][-self.history_horizon:]

    def update(self):
        """Handle updated solution results"""
        try:
            # Load options.json
            self.options = json.loads(open(os.path.join(self.path,'options.json')).read())
            # Load results.json
            self.results = json.loads(open(os.path.join(self.path,"results.json")).read())
        except Exception as e:
            raise e
        # Updates the options_dict
        self.get_options()

        # Make sure we don't overflow the history_horizon
        self.check_history_horizon()

        time = self.gekko_data['time']
        if len(time) > 1:
            time_interval = time[1] - time[0]

        # Update the time array here
        if self.options['APM']['IMODE'] in (5, 6, 8, 9):
            if len(time) > 1:
                self.gekko_data['time'].append(time[-1] + time_interval)

        # Append new vars data here
        main_dict = vars(main)
        for var in main_dict:
            if (var != 'time') and isinstance(main_dict[var], (GKVariable, GKParameter, GK_Intermediate)):
                self.get_var_from_main(var)

        # Let the GUI know the updates are ready
        self.has_new_update = True


class GK_GUI:
    """GUI class for GEKKO
    Creates and manages the FlaskThread that actually runs the API.
    """
    def __init__(self, path):
        self.path = path

    def display(self):
        """Finds the appropriate port starts the api and opens the webbrowser"""
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # It is necessary to have a reference like this as flaskThread cannot
        # be part of self when `flaskThread.start()` is called for reasons I
        # do not fully understand. Improve on this if you can.
        self.apiRef = "This is a reference to the Flask API thread"
        port = 8050
        try:    # Check to see if :8050 is already bound
            sock.bind(('127.0.0.1', port))
            sock.close()
        except OSError:   # Find an open port if :8050 is taken
            sock.bind(('127.0.0.1', 0))
            port = sock.getsockname()[1]
            sock.close()
        except socket.error:
            sock.bind(('127.0.0.1', 0))
            port = sock.getsockname()[1]
            sock.close()

        # Open the browser to the page and launch the app
        print('Opening display in default webbrowser at http://localhost:' + str(port) + '/index.html. \nClose display tab or type CTRL+C to exit.')
        if DEV:
            print('Starting Flask Thread on port {}'.format(8050))
            flaskThread = FlaskThread(self.path, True, 8050)
            # flaskThread.daemon = True
            flaskThread.start()
            self.apiRef = flaskThread
        else:
            print('Starting Flask Thread on port {}'.format(port))
            webbrowser.open("http://localhost:" + str(port) + "/index.html")
            flaskThread = FlaskThread(self.path, False, port)
            # flaskThread.daemon = True
            flaskThread.start()
            self.apiRef = flaskThread
            # Non-threaded way, only works for non-dynamic GUI display
            # app.run(debug=False, port=port)

    def update(self):
        """Alert the API of new solution results"""
        if DEV:
            print('Handling update')
        self.apiRef.update()

    



# -*- coding: utf-8 -*-
import numpy as np
"""This class enables operator overloading for turning python equations into
strings for the .apm model. Each variable type inherits this class. Operations 
done on an instance of this class return a new instance to enable chained 
operations."""
class GK_Operators:
    """"""
    count = 0
    
    def __init__(self, name, value=None):                
        if name is None:
            self.NAME = 'i'+str(GK_Operators.count)
            GK_Operators.count += 1
        else:
            self.NAME = name
        
        self.VALUE = GK_Value(value)
            
    def __repr__(self):
        return self.name
    def __str__(self):
        return self.name
    def __len__(self):
        return len(self.value)
    def __getitem__(self,key):
        return self.value[key]
    #make attributes case in-sensitive for reading too
    # (this is inherited by variables and paramters)
    def __getattr__(self,name):
        if name.upper() in self.__dict__:
            return self.__dict__[name.upper()]
        elif name.lower() in self.__dict__:
            return self.__dict__[name.lower()]
        else: 
            raise AttributeError(name)
    #%%Operator overloading for building functions
    #comparisons
    def __lt__(self,other): #less than
        return GK_Operators(str(self) + '<' + str(other))
    def __le__(self,other): #less than or equal to
        return GK_Operators(str(self) + '<=' + str(other))
    def __gt__(self,other): #greater than
        return GK_Operators(str(self) + '>' + str(other))
    def __ge__(self,other): #greater than or equal to
        return GK_Operators(str(self) + '>=' + str(other))
    def __eq__(self,other): #equal ==
        return GK_Operators(str(self) + '=' + str(other))
    #math operators
    def __add__(self,other): # +
        return GK_Operators('(' + str(self) + '+' + str(other) + ')')
    def __sub__(self,other): # -
        return GK_Operators('(' + str(self) + '-' + str(other) + ')')
    def __pow__(self,other): # **
        return GK_Operators('((' + str(self) + ')^(' + str(other) + '))')
    def __div__(self,other): # /
        return GK_Operators('((' + str(self) + ')/(' + str(other) + '))')
    def __truediv__(self,other): # /
        return GK_Operators('((' + str(self) + ')/(' + str(other) + '))')
    def __mul__(self,other): # *
        return GK_Operators('((' + str(self) + ')*(' + str(other) + '))')
    def __neg__(self): #-x
        return GK_Operators('(-'+str(self)+')')
    # reverse math    
    def __radd__(self,other): # +
        return GK_Operators('(' + str(other) + '+' + str(self) + ')')
    def __rsub__(self,other): # -
        return GK_Operators('(' + str(other) + '-' + str(self) + ')')
    def __rpow__(self,other): # **
        return GK_Operators('(' + str(other) + '^' + str(self) + ')')
    def __rdiv__(self,other): # /
        return GK_Operators('(' + str(other) + '/' + str(self) + ')')
    def __rtruediv__(self,other): # /
        return GK_Operators('(' + str(other) + '/' + str(self) + ')')
    def __rmul__(self,other): # *
        return GK_Operators('((' + str(other) + ')*(' + str(self) + '))')
    #other
    def __abs__(self):
        return GK_Operators('abs('+str(self)+')')
    """
    object.__iadd__(self, other)
    object.__isub__(self, other)
    object.__imul__(self, other)
    object.__imatmul__(self, other)
    object.__itruediv__(self, other)
    object.__ifloordiv__(self, other)
    object.__imod__(self, other)
    object.__ipow__(self, other[, modulo])
    object.__ilshift__(self, other)
    object.__irshift__(self, other)
    object.__iand__(self, other)
    object.__ixor__(self, other)
    object.__ior__(self, other)
    object.__neg__(self)
    object.__pos__(self)
    object.__abs__(self)
    
    These work in principle but python ranks ^ quite low in the order of operations
    so it gets all funky
    def __xor__(self,other): # ^
        return GK_Operators('(' + str(self) + '^' + str(other) + ')')
    def __rxor__(self,other): # ^
        return GK_Operators('(' + str(other) + '^' + str(self) + ')')
    """

class GK_Intermediate(GK_Operators):
    def __init__(self, name, value=None):
        GK_Operators.__init__(self,name, value=None)
        
    def __repr__(self):
        return str(self.value) #'%s = %f' % (self.name, self.value)
    def __len__(self):
        return len(self.value)
    def __getitem__(self,key):
        return self.value[key]
     
      
class GK_Value(list):
    def __init__(self,value):
        if value is not None:
            self.change = True
            try:
                # value is a SV, CV, Var, Param
                value = value.value.value
            except:
                try:
                    # value is a GK_Value
                    value = value.value
                except:
                    # value is a number
                    pass
            self.value = value
        else:
            self.value = 0 #store default value without triggering change detection
            self.change = False
            
    def __repr__(self):
        return str(self.value)
    def __str__(self):
        return str(self.value)
    
    
    def __len__(self):
        return len(self.value)
    
    def __getitem__(self,key):
        return self.value[key]
    
    def __setattr__(self, name, value):
        if name == 'value':
            self.__dict__[name] = value
            self.__dict__['change'] = True
        elif name == 'change':
            self.__dict__['change'] = value
        else:
            raise Exception('Unrecognized property')
    
    def __setitem__(self,key,value):
        self.value[key] = value
        if self.change is False:
            self.__dict__['change'] = [key]
        elif isinstance(self.change,list): 
            self.__dict__['change'].append(key)
        else:
            pass
    
    def __array__(self):
        return np.array(self.value)

    def __iter__(self):
        try:
            for v in self.value:
                yield v
        except:
            yield self.value
                  
    #%%Operator overloading for building functions
    #comparisons
    def __lt__(self,other): #less than
        return self.value < other
    def __le__(self,other): #less than or equal to
        return self.value <= other
    def __gt__(self,other): #greater than
        return self.value > other
    def __ge__(self,other): #greater than or equal to
        return self.value >= other
    def __eq__(self,other): #equal ==
        return self.value == other
    #math operators
    def __add__(self,other): # +
        return self.value + other
    def __sub__(self,other): # -
        return self.value - other
    def __pow__(self,other): # **
        return self.value ** other
    def __div__(self,other): # /
        return self.value / other
    def __truediv__(self,other): # /
        return self.value / other
    def __mul__(self,other): # *
        return self.value * other
    def __neg__(self): #-x
        return self.value*-1
    # reverse math    
    def __radd__(self,other): # +
        return other + self.value
    def __rsub__(self,other): # -
        return other - self.value
    def __rpow__(self,other): # **
        return other ** self.value
    def __rdiv__(self,other): # /
        return other / self.value
    def __rtruediv__(self,other): # /
        return other / self.value
    def __rmul__(self,other): # *
        return other * self.value
    #other
    def __abs__(self):
        return abs(self.value)



import os 
from .gk_operators import GK_Operators
"""
Param_input_options = []
Param_inout_options = ['VALUE']
Param_output_options = []


FV_input_options = Param_input_options+['CRITICAL', 'DMAX', 'DMAXHI', 
                                        'DMAXLO', 'FSTATUS', 'LOWER','MEAS', 
                                        'PSTATUS','STATUS', 'UPPER', 'VDVL', 
                                        'VLACTION', 'VLHI', 'VLLO']
FV_inout_options = Param_inout_options+[]
FV_output_options = Param_output_options+['LSTVAL', 'NEWVAL']

MV_inout_options = FV_inout_options+[]
MV_input_options = FV_input_options + ['COST', 'DCOST', 'MV_STEP_HOR','REQONCTRL', 'TIER']
MV_output_options = FV_output_options + ['AWS', 'DPRED', 'NXTVAL', 'PRED', ]

#gather in dictionary
parameter_options = {'FV':{'inputs':FV_input_options, 'outputs':FV_output_options, 'inout': FV_inout_options}, 
                     'MV':{'inputs':MV_input_options,'outputs':MV_output_options,'inout':MV_inout_options},
                     None:{'inputs':Param_input_options,'outputs':Param_output_options,'inout':Param_inout_options}}

"""
from .properties import parameter_options as options

class GKParameter(GK_Operators):
    """Represents a parameter in a model."""
    counter = 1

    def __init__(self, name='', value=None, lb=None, ub=None, integer=False):
        if name == '':
            name = 'p' + GKParameter.counter
            GKParameter.counter += 1
            if integer == True:
                name = 'int_' + name
                
        # prevents the __setattr__ function from sending options to the server
        # until the __init__ function has completed since they should only be
        # sent if changed from their defaults
        self.__dict__['_initialized'] = False
        
        #register fixed values through connections to ensure consistency in the 
        #csv file, otherwise the requested fixed value will be overridden by
        #whatever initialization value is in the csv
        #self._override_csv = []        
        self._override_csv = []
                
        GK_Operators.__init__(self, name, value=value)

        #self.VALUE = value #initialized value SET IN GK_Operators
            
        if not hasattr(self,'type'): #don't overwrite FV and MV
            self.type = None 
       
        # parameters can have lower and upper bounds       
        if lb is not None:
            self.LOWER = lb
        else:
            self.LOWER = None
        if ub is not None:
            self.UPPER = ub
        else:
            self.UPPER = None
        
        # now allow options to be sent to the server
        self._initialized = True
        
        
    def __repr__(self):
        return str(self.value) #'%s = %f' % (self.name, self.value)

    def __len__(self):
        return len(self.value)
    def __getitem__(self,key):
        return self.value[key]
    def __setitem__(self,key,value):
        self.value[key] = value

#    def __getattr__(self,name):
#        name = name.upper()
#        if name == 'VALUE':
#            return self.__dict__['VALUE'].value
#        else:
#            return self.__dict__[name]
    
    def __setattr__(self, name, value):
        if self._initialized:
            #ignore cases on global options
            name = name.upper()

            #only allow user to set input or input/output options:
            if name in options[self.type]['inputs']+options[self.type]['inout']:
                if name == 'VALUE':
                    # Extract input array from pandas series if needed
                    if type(value).__name__ == 'Series':
                        value = value.values
                    self.__dict__[name].value = value
                else:
                    self.__dict__[name] = value

                    
            #don't allow writing to output properties by default
            elif name in options[self.type]['outputs']:
                #define outputs by passing list/tuple with 1st element being True
                #to override the output writing prevention 
                try:
                    if value[0] == True:
                        self.__dict__[name] = value[1]
                    else:
                        raise TypeError
                except TypeError:
                    raise AttributeError(str(name)+" is an output property")

                    
            #no other properties allowed
            else:
                raise AttributeError(str(name)+" is not a property of this variable")

                
        #for initializing model
        else:
            self.__dict__[name] = value
            
            
class GK_FV(GKParameter):
    """Fixed Variable. Inherits GKParameter."""

    def __init__(self, name='', value=0, lb=None, ub=None, gk_model=None, model_path=None, integer=False):

        # prevents the __setattr__ function from sending options to the server
        # until the __init__ function has completed since they should only be
        # sent if changed from their defaults
        self.__dict__['_initialized'] = False

        if not hasattr(self,'type'): #don't overwrite MV
            self.type = 'FV'
        self.model_name = gk_model
        self.path = model_path #use the same path as the model 
        
        # FV options
        self.CRITICAL = None
        self.DMAX = None
        self.DMAXHI = None
        self.DMAXLO = None
        self.FSTATUS = None
        self.LSTVAL = None
        self.MEAS = None
        self.NEWVAL = None
        self.PSTATUS = None
        self.STATUS = None
        self.VDVL = None
        self.VLACTION = None
        self.VLHI = None
        self.VLLO = None
       
        GKParameter.__init__(self, name=name, value=value, lb=lb, ub=ub, integer=integer)


    def meas(self,measurement):
        self.MEAS = measurement
        #open measurement.dbs file
        f = open(os.path.join(self.path,'measurements.dbs'),'a')
        #write measurement
        f.write(self.name+'.MEAS = '+str(measurement)+', 1, none\n')
        #close file
        f.close()
        
        #write tag file
        f = open(os.path.join(self.path,self.name),'w')
        #write measurement
        f.write(str(measurement))
        #close tag file
        f.close()
        
        
        



class GK_MV(GK_FV):
    """ Manipulated Variable. Inherits GK_FV."""

    def __init__(self, name='', value=0, lb=None, ub=None, gk_model=None, model_path=None, integer=False):
        
        # prevents the __setattr__ function from sending options to the server
        # until the __init__ function has completed since they should only be
        # sent if changed from their defaults
        self.__dict__['_initialized'] = False

        # prevents the __setattr__ function from sending options to the server
        # until the __init__ function has completed since they should only be
        # sent if changed from their defaults
        self.initialized = False

        self.type = 'MV'

        # options for manipulated variables
        self.AWS = None
        self.COST = None
        self.DCOST = None
        self.DPRED = None
        self.MV_STEP_HOR = None
        self.NXTVAL = None
        self.PRED = None
        self.REQONCTRL = None
        self.TIER = None
        
        
        GK_FV.__init__(self, name=name, value=value, lb=lb, ub=ub, gk_model=gk_model, model_path=model_path, integer=integer)

        



# -*- coding: utf-8 -*-
from gekko import GEKKO
import numpy as np
"""
GEKKO specializes in a unique subset of machine learning. However, it can be used
for various types of machine learning. This is a module to facilitate Artificial 
Neural Networks in GEKKO.

Most ANN packages uses gradient decent optimization. The solvers used in GEKKO 
use far more advanced techniques that gradient decent. However, training neural 
networks requires extremely large datasets. For these large problems, gradient 
decent does prove more useful because of its ability to massively parallelize.
Nevertheless, training in gekko is available.
"""


class Brain():
    
    def __init__(self,m=[],remote=True,bfgs=True,explicit=True):
        if m==[]:
            self.m = GEKKO(remote=remote)
        else:
            self.m = m
        #generic model options
        self.m.options.MAX_ITER = 4000
        self.m.options.OTOL = 1e-4
        self.m.options.RTOL = 1e-4
        if bfgs:
            self.m.solver_options = ['hessian_approximation limited-memory']
        
        self._explicit = explicit 
        self._input_size = None
        self._output_size = None
        self._layers = []
        self._weights = []
        self._biases = []
        self.input = []
        self.output = []
        
        
    def input_layer(self,size):
        #store input size
        self._input_size = size
        #build FV with Feedback to accept inputs
        self.input = [self.m.Param() for _ in range(size)]
#        #set FV options
#        for n in self.input:
#            n.FSTATUS = 1
#            n.STATUS = 0
            
        #add input layer to list of layers
        self._layers.append(self.input)

    def layer(self,linear=0,relu=0,tanh=0,gaussian=0,bent=0,leaky=0,ltype='dense'):
        """
        Layer types:
            dense
            convolution
            pool (mean)
        
        Activation options:
            none
            softmax
            relu
            tanh
            sigmoid
            linear
        """
        size = relu + tanh + linear + gaussian + bent + leaky
        if size < 1:
            raise Exception("Need at least one node")
        
        if ltype == 'dense':
            
            ## weights between neurons
            n_p = len(self._layers[-1]) #number of neuron in previous layer
            n_c = n_p * size # number of axion connections
            # build n_c FVs as axion weights, initialize randomly in [-1,1]
            self._weights.append([self.m.FV(value=[np.random.rand()*2-1]) for _ in range(n_c)])
            for w in self._weights[-1]:
                w.STATUS = 1
                w.FSTATUS = 0
            #input times weights, add bias and activate
            self._biases.append([self.m.FV(value=0) for _ in range(size)])
            for b in self._biases[-1]:
                b.STATUS = 1
                b.FSTATUS = 0
            
            
            count = 0
            
            if self._explicit:
            
                # build new neuron weighted inputs
                #i counts nodes in this layer, j counts nodes of previous layer
                neuron_inputs = [self.m.Intermediate(self._biases[-1][i] \
                                                     + sum((self._weights[-1][(i*n_p)+j]*self._layers[-1][j]) \
                                                           for j in range(n_p))) for i in range(size)] 
                
                ##neuron activation
                self._layers.append([])
                if linear > 0:
                    self._layers[-1] += [self.m.Intermediate(neuron_inputs[i]) \
                                         for i in range(count,count+linear)]
                    count += linear
                if tanh > 0:
                    self._layers[-1] += [self.m.Intermediate(self.m.tanh(neuron_inputs[i])) \
                                         for i in range(count,count+tanh)]
                    count += tanh
                if relu > 0:
                    self._layers[-1] += [self.m.Intermediate(self.m.log(1+self.m.exp(neuron_inputs[i]))) \
                                         for i in range(count,count+relu)]
                    count += relu
                if gaussian > 0:
                    self._layers[-1] += [self.m.Intermediate(self.m.exp(-neuron_inputs[i]**2)) \
                                         for i in range(count,count+gaussian)]
                    count += gaussian
                if bent > 0:
                    self._layers[-1] += [self.m.Intermediate((self.m.sqrt(neuron_inputs[i]**2 + 1) - 1)/2 \
                                                             + neuron_inputs[i]) \
                                         for i in range(count,count+bent)]
                    count += bent
                if leaky > 0:
                    s = [self.m.Var(lb=0) for _ in range(leaky*2)]
                    self.m.Equations( [ (1.5*neuron_inputs[i+count]) \
                                        - (0.5*neuron_inputs[i+count]) \
                                        == s[2*i] - s[2*i+1] for i in range(leaky) ] )
                    self._layers[-1] += [self.m.Intermediate(neuron_inputs[count+i] + s[2*i]) \
                                         for i in range(leaky)]
                    [self.m.Obj(s[2*i]*s[2*i+1]) for i in range(leaky)]
                    self.m.Equations([s[2*i]*s[2*i+1] == 0 for i in range(leaky)])
                    count += leaky
                    
            
            else: #type=implicit
                 
                # build new neuron weighted inputs
                neuron_inputs = [self.m.Var() for i in range(size)]
                #i counts nodes in this layer, j counts nodes of previous layer
                self.m.Equations([neuron_inputs[i] == self._biases[-1][i] \
                                  + sum((self._weights[-1][(i*n_p)+j]*self._layers[-1][j]) \
                                        for j in range(n_p)) for i in range(size)]) 
                
                ##neuron activation
                neurons = [self.m.Var() for i in range(size)]
                self._layers.append(neurons)
                
                ##neuron activation
                if linear > 0:
                    self.m.Equations([neurons[i] == neuron_inputs[i] \
                                      for i in range(count,count+linear)])
                    count += linear
                if tanh > 0:
                    self.m.Equations([neurons[i] == self.m.tanh(neuron_inputs[i]) \
                                      for i in range(count,count+tanh)])
                    for n in neurons[count:count+tanh]:
                        n.LOWER = -5
                        n.UPPER = 5
                    count += tanh
                if relu > 0:
                    self.m.Equations([neurons[i] == self.m.log(1+self.m.exp(neuron_inputs[i])) \
                                      for i in range(count,count+relu)])
                    for n in neurons[count:count+relu]:
                        n.LOWER = -10 
                    count += relu
                if gaussian > 0:
                    self.m.Equations([neurons[i] == self.m.exp(-neuron_inputs[i]**2) \
                                      for i in range(count,count+gaussian)])
                    for n in neurons[count:count+gaussian]:
                        n.LOWER = -3.5
                        n.UPPER = 3.5
                    count += gaussian
                if bent > 0:
                    self.m.Equations([neurons[i] == ((self.m.sqrt(neuron_inputs[i]**2 + 1) - 1)/2 \
                                                     + neuron_inputs[i]) for i in range(count,count+bent)])
                    count += bent
                if leaky > 0:
                    s = [self.m.Var(lb=0) for _ in range(leaky*2)]
                    self.m.Equations( [ (1.5*neuron_inputs[count+i]) - (0.5*neuron_inputs[count+i]) \
                                        == s[2*i] - s[2*i+1] for i in range(leaky) ] )
                    self.m.Equations([neurons[count+i] == neuron_inputs[count+i] + s[2*i] \
                                      for i in range(leaky)])
                    [self.m.Obj(10000*s[2*i]*s[2*i+1]) for i in range(leaky)]
                    #self.m.Equations([s[2*i]*s[2*i+1] == 0 for i in range(leaky)])
                    count += leaky
                    
                    
        else:
            raise Exception('layer type not implemented yet')
                
            
    
    
    def output_layer(self,size,ltype='dense',activation='linear'):
        """
        Layer types:
            dense
            convolution
            pool (mean)
        
        Activation options:
            none
            softmax
            relu
            tanh
            sigmoid
            linear
        """
        
        # build a layer to ensure that the number of nodes matches the output
        self.layer(size,0,0,0,0,0,ltype)
        
        self.output = [self.m.CV() for _ in range(size)]
        for o in self.output:
            o.FSTATUS = 1
            o.STATUS = 1
        
        #link output CVs to last layer
        for i in range(size):
            self.m.Equation(self.output[i] == self._layers[-1][i])

    def think(self,inputs):
        
        #convert inputs to numpy ndarray
        inputs = np.atleast_2d(inputs)
        
        ##confirm input/output dimensions
        in_dims = inputs.shape
        ni = len(self.input)
        #consistent layer size
        if in_dims[0] != ni:
            raise Exception('Inconsistent number of inputs')
        
        #set input values
        for i in range(ni):
            self.input[i].value = inputs[i,:]
    
        #solve in SS simulation
        self.m.options.IMODE = 2
        #disable all weights
        for wl in self._weights:
            for w in wl:
                w.STATUS = 0
        for bl in self._biases:
            for b in bl:
                b.STATUS = 0
        self.m.solve(disp=False)
        
        ##return result
        res = [] #concatentate result from each CV in one list
        for i in range(len(self.output)):
            res.append(self.output[i].value)
        
        return res    
        

    def learn(self,inputs,outputs,obj=2,gap=0,disp=True):
        """
        Make the brain learn. 
        Give inputs as (n)xm
            Where n = input layer dimensions
            m = number of datasets
        Give outputs as (n)xm
            Where n = output layer dimensions
            m = number of datasets
        Objective can be 1 (L1 norm) or 2 (L2 norm)
        If obj=1, gap provides a deadband around output matching.
        """
        
        #convert inputs to numpy ndarray
        inputs = np.atleast_2d(inputs)
        outputs = np.atleast_2d(outputs)
        
        ##confirm input/output dimensions
        in_dims = inputs.shape
        out_dims = outputs.shape
        ni = len(self.input)
        no = len(self.output)
        #consistent dataset size
        if in_dims[1] != out_dims[1]:
            raise Exception('Inconsistent number of datasets')
        #consistent layer size
        if in_dims[0] != ni:
            raise Exception('Inconsistent number of inputs')
        if out_dims[0] != no:
            raise Exception('Inconsistent number of outputs')
        
        #set input values
        for i in range(ni):
            self.input[i].value = inputs[i,:]
        
        #set output values
        for i in range(no):
            o = self.output[i]
            o.value = outputs[i,:]
            if obj == 1: #set meas_gap while cycling through CVs
                o.MEAS_GAP = gap
            
        #solve in MPU mode
        self.m.options.IMODE = 2
        self.m.options.EV_TYPE = obj
        self.m.options.REDUCE = 3
        #enable all weights
        for wl in self._weights:
            for w in wl:
                w.STATUS = 1
        for bl in self._biases:
            for b in bl:
                b.STATUS = 1
            
        self.m.solve(disp=disp)
        
    def shake(self,percent):
        """ Neural networks are non-convex. Some stochastic shaking can 
        sometimes help bump the problem to a new region. This function 
        perturbs all weights by +/-percent their values."""
        
        for l in self._weights:
            for f in l:
                f.value = f.value[-1]*(1+(1-2*np.random.rand())*percent/100)





# -*- coding: utf-8 -*-
"""
Created on Fri Dec 15 12:17:19 2017

@author: scd
"""
#%% global model options
"""
global_option_list = ['APPINFO', 'APPINFOCHG', 'APPSTATUS', 'BNDS_CHK', 'COLDSTART',
                      'CSV_READ', 'CSV_WRITE', 'CTRLMODE', 'CTRL_HOR', 'CTRL_TIME',
                      'CTRL_UNITS', 'CV_TYPE', 'CV_WGT_SLOPE', 'CV_WGT_START',
                      'CYCLECOUNT', 'DBS_READ', 'DBS_WRITE', 'DIAGLEVEL', 'EV_TYPE',
                      'EV_WGT_SLOPE', 'HIST_HOR', 'HIST_UNITS', 'ICD_CALC', 'IMODE',
                      'ITERATIONS', 'MAX_ITER', 'MAX_TIME', 'MEAS_CHK', 'MV_DCOST_SLOPE',
                      'MV_STEP_HOR', 'MV_TYPE', 'NODES', 'OBJFCNVAL', 'OTOL', 'PRED_HOR',
                      'PRED_TIME', 'REDUCE', 'REQCTRLMODE', 'RTOL', 'SCALING', 'SENSITIVITY',
                      'SOLVESTATUS', 'SOLVER', 'SOLVETIME', 'SPECS', 'TIME_SHIFT',
                      'WEB', 'WEB_MENU', 'WEB_REFRESH']
"""                      
global_options_inputs =  ['AUTO_COLD', 'BNDS_CHK', 'CSV_READ', 'CSV_WRITE', 
'CTRL_UNITS', 'CV_WGT_SLOPE', 'CV_WGT_START', 'CV_TYPE', 'DBS_LEVEL', 
'DBS_READ', 'DBS_WRITE', 'DIAGLEVEL', 'EV_WGT_SLOPE', 'EV_TYPE', 'FILTER', 
'FRZE_CHK', 'HIST_HOR', 'HIST_UNITS', 'ICD_CALC', 'IMODE', 'LINEAR', 'MAX_ITER', 
'MAX_MEMORY', 'MAX_TIME', 'MEAS_CHK', 'MV_DCOST_SLOPE', 'MV_STEP_HOR', 
'MV_TYPE', 'NODES', 'OTOL', 'REDUCE', 'REPLAY', 'REQCTRLMODE', 'RTOL', 
'SCALING', 'SENSITIVITY', 'SEQUENTIAL', 'SOLVER', 'SPECS', 'SPC_CHART', 
'STREAM_LEVEL', 'TIME_SHIFT', 'WEB', 'WEB_MENU', 'WEB_REFRESH', 'WEB_PLOT_FREQ']

global_options_outputs= ['APPINFO', 'APPINFOCHG', 'APPSTATUS', 'CTRLMODE', 
'ITERATIONS', 'OBJFCNVAL', 'SOLVESTATUS', 'SOLVETIME']

global_options_inout = ['BAD_CYCLES', 'COLDSTART', 'CTRL_HOR', 'CTRL_TIME', 
'CYCLECOUNT', 'PRED_HOR', 'PRED_TIME']

#gather in dictionary
global_options = {'inputs':global_options_inputs,'outputs':global_options_outputs,'inout':global_options_inout}
#%% Parameter options

# TODO: check if OSTATUS, OSTATUSCHG, PRED, DPRED, AWS, VLACTION, TIER are correctly exposed

Param_input_options = ['LOWER', 'UPPER']
Param_inout_options = ['VALUE']
Param_output_options = []


FV_input_options = Param_input_options+['CRITICAL', 'DMAX', 'DMAXHI', 
                                        'DMAXLO', 'FSTATUS', 'LOWER','MEAS', 
                                        'PSTATUS','STATUS', 'UPPER', 'VDVL', 
                                        'VLACTION', 'VLHI', 'VLLO']
FV_inout_options = Param_inout_options+[]
FV_output_options = Param_output_options+['LSTVAL', 'NEWVAL']

MV_input_options = FV_input_options + ['COST', 'DCOST', 'MV_STEP_HOR','REQONCTRL', 'TIER']
MV_inout_options = FV_inout_options+[]
MV_output_options = FV_output_options + ['AWS', 'DPRED', 'NXTVAL', 'PRED', ]

#gather in dictionary
parameter_options = {'FV':{'inputs':FV_input_options, 'outputs':FV_output_options, 'inout': FV_inout_options}, 
                     'MV':{'inputs':MV_input_options,'outputs':MV_output_options,'inout':MV_inout_options},
                     None:{'inputs':Param_input_options,'outputs':Param_output_options,'inout':Param_inout_options}}



#%% Variable options
Var_input_options = ['LOWER', 'UPPER']
Var_inout_options = ['VALUE']
Var_output_options = []

SV_input_options = Var_input_options+['FSTATUS', 'MEAS']
SV_inout_options = Var_inout_options+[]
SV_output_options = Var_output_options+['MODEL', 'PRED']

CV_inout_options = SV_inout_options+['BIAS']
CV_input_options = SV_input_options + ['COST', 'CRITICAL', 'FDELAY', 
                                       'MEAS_GAP', 'PSTATUS', 'SP', 'SPHI', 
                                       'SPLO', 'STATUS', 'TAU', 'TIER', 'TR_INIT', 
                                       'TR_OPEN', 'VDVL', 'VLACTION', 'VLHI',
                                       'VLLO', 'WMEAS', 'WMODEL', 'WSP', 
                                       'WSPHI', 'WSPLO']
CV_output_options = SV_output_options + ['LSTVAL']

#gather in dictionary
variable_options = {'SV':{'inputs':SV_input_options, 'outputs':SV_output_options, 'inout': SV_inout_options}, 
                    'CV':{'inputs':CV_input_options,'outputs':CV_output_options,'inout':CV_inout_options},
                    None:{'inputs':Var_input_options,'outputs':Var_output_options,'inout':Var_inout_options}}



# -*- coding: utf-8 -*-

import json
import os

from .properties import parameter_options, variable_options


#%% Post-solve processing

## options.JSON has all APM options
def load_JSON(self):
    f = open(os.path.join(self._path,'options.json'))
    data = json.load(f)
    f.close()
    #global (APM) options
    for o in self.options._output_option_list+self.options._inout_option_list:
        self.options.__dict__[o] = data['APM'][o]
    #Variable options (FV/MV/SV/CV)
    for vp in self._parameters:
        if vp.type != None: #(FV/MV/SV/CV) not Param or Var
            for o in parameter_options[vp.type]['outputs']+parameter_options[vp.type]['inout']:
                if o == 'VALUE':
                    continue
                elif o == 'PRED': #Pred can be an array of up to 10
                    if o in data[vp.name]: #single value
                        vp.__dict__[o] = data[vp.name][o]
                    else:
                        try: #fill in an array up to 10 values
                            pred = []
                            for i in range(11):
                                pred.append(data[vp.name][o+'['+str(i)+']'])
                        except:
                            pass
                        finally:
                            vp.__dict__[o] = pred
                elif o == 'DPRED': #Pred can be an array of up to 10
                    if o in data[vp.name]: #single value
                        vp.__dict__[o] = data[vp.name][o]
                    else:
                        try: #fill in an array up to 10 values
                            dpred = []
                            for i in range(1,11):
                                pred.append(data[vp.name][o+'['+str(i)+']'])
                        except:
                            pass
                        finally:
                            vp.__dict__[o] = dpred
                else: #everything besides value, dpred and pred
                    vp.__dict__[o] = data[vp.name][o]
    for vp in self._variables:
        if vp.type != None: #(FV/MV/SV/CV) not Param or Var
            for o in variable_options[vp.type]['outputs']+variable_options[vp.type]['inout']:

                if o == 'VALUE':
                    continue
                elif o == 'PRED': #Pred can be an array of up to 10
                    if o in data[vp.name]: #single value
                        vp.__dict__[o] = data[vp.name][o]
                    else:
                        try: #fill in an array up to 10 values
                            pred = []
                            for i in range(11):
                                pred.append(data[vp.name][o+'['+str(i)+']'])
                        except:
                            pass
                        finally:
                            vp.__dict__[o] = pred
                else: #everything besides value and pred
                    vp.__dict__[o] = data[vp.name][o]
    return data


## results.json has variable value results
def load_results(self):
    if (os.path.isfile(os.path.join(self._path, 'results.json'))):
        f = open(os.path.join(self._path,'results.json'))
        data = json.load(f)
        f.close()

        for vp in self._parameters:
            try:
                vp.VALUE = data[vp.name]
                vp.value.change = False
            except Exception:
                print(vp.name+ " not found in results file")
        for i in self._intermediates:
            try:
                i.value.value = data[i.name]
                i.value.change = False
            except Exception:
                print(i.name+ " not found in results file")
        for vp in self._variables:
            try:
                vp.VALUE = data[vp.name]
                vp.value.change = False
            except Exception:
                print(vp.name+ " not found in results file")

        return data

    else:
        print("Error: 'results.json' not found. Check above for additional error details")
        return {}

    print(data['APM']['SOLVESTATUS'])
    return data





import os
from .gk_operators import GK_Operators


"""
Var_input_options = []
Var_inout_options = ['VALUE']
Var_output_options = []

SV_input_options = Var_input_options+['FSTATUS', 'LOWER', 'MEAS', 'UPPER']
SV_inout_options = Var_inout_options+[]
SV_output_options = Var_output_options+['MODEL', 'PRED']

CV_inout_options = SV_inout_options+['BIAS']
CV_input_options = SV_input_options + ['COST', 'CRITICAL', 'FDELAY', 
                                       'MEAS_GAP', 'PSTATUS', 'SP', 'SPHI', 
                                       'SPLO', 'STATUS', 'TAU', 'TIER', 'TR_INIT', 
                                       'TR_OPEN', 'VDVL', 'VLACTION', 'VLHI',
                                       'VLLO', 'WMEAS', 'WMODEL', 'WSP', 
                                       'WSPHI', 'WSPLO']
CV_output_options = SV_output_options + ['LSTVAL' ]

#gather in dictionary
variable_options = {'SV':{'inputs':SV_input_options, 'outputs':SV_output_options, 'inout': SV_inout_options}, 
                    'CV':{'inputs':CV_input_options,'outputs':CV_output_options,'inout':CV_inout_options},
                    None:{'inputs':Var_input_options,'outputs':Var_output_options,'inout':Var_inout_options}}"""
from .properties import variable_options as options

class GKVariable(GK_Operators):
    """Represents a parameter in a model"""
    counter = 0
    
    def __init__(self, name='', value=None, lb=None, ub=None, integer=False):
        if name == '':
            name = 'v' + GKVariable.counter
            GKVariable.counter += 1
            if integer == True:
                name = 'int_'+name
        
        
        # prevents the __setattr__ function from sending options to the server
        # until the __init__ function has completed since they should only be
        # sent if changed from their defaults
        self.__dict__['_initialized'] = False

        GK_Operators.__init__(self, name, value=value)

        #self.VALUE = value #initialized value is done in GK_Operators
        if not hasattr(self,'type'): #don't overwrite SV and CV
            self.type = None 
            
        if lb is not None:
            self.LOWER = lb
        else:
            self.LOWER = None
        if ub is not None:
            self.UPPER = ub
        else:
            self.UPPER = None
        
        #register fixed values through connections to ensure consistency in the 
        #csv file, otherwise the requested fixed value will be overridden by
        #whatever initialization value is in the csv
        self._override_csv = []
        
        #register values that are changed by the user 
        #self._changed = True
        # now allow options to be sent to the server
        self._initialized = True

    def dt(self):
        return GK_Operators('$' + self.name)

    def __repr__(self):
        return str(self.value)
    
    def __len__(self):
        return len(self.value)
    def __getitem__(self,key):
        return self.value[key]
    def __setitem__(self,key,value):
        self.value[key] = value


#    def __getattr__(self,name):
#        name = name.upper()
#        if name == 'VALUE':
#            return self.__dict__['VALUE'].value
#        else:
#            return self.__dict__[name]

    def __setattr__(self, name, value):
        if self._initialized:
            #ignore cases on global options
            name = name.upper()

            #only allow user to set input or input/output options:
            if name in options[self.type]['inputs']+options[self.type]['inout']:
                if name == 'VALUE':
                    # Extract input array from pandas series if needed
                    if type(value).__name__ == 'Series':
                        value = value.values
                    self.__dict__[name].value = value
                else:
                    self.__dict__[name] = value
                    
                        
            #don't allow writing to output properties by default
            elif name in options[self.type]['outputs']:
                #define outputs by passing list/tuple with 1st element being True
                #to override the output writing prevention 
                try:
                    if value[0] == True:
                        self.__dict__[name] = value[1]
                    else:
                        raise TypeError
                except TypeError:
                    raise AttributeError(str(name)+" is an output property")

                    
            #no other properties allowed
            else:
                raise AttributeError(str(name)+" is not a property of this variable")
                
        #for initializing model
        else:
            self.__dict__[name] = value
        



class GK_SV(GKVariable):
    """State Variable. Inherits GKVariable."""

    def __init__(self, name='', value=0, lb=None, ub=None, gk_model=None, model_path=None, integer=False):

        # prevents the __setattr__ function from sending options to the server
        # until the __init__ function has completed since they should only be
        # sent if changed from their defaults
        self.__dict__['_initialized'] = False
        
        if not hasattr(self,'type'): #don't overwrite CV
            self.type = 'SV'
        self.model_name = gk_model 
        self.path = model_path #use the same path as the model 
        
        # SV specific options
        self.FSTATUS = None
        self.LOWER = None
        self.MEAS = None
        self.MODEL = None
        self.PRED = None
        self.UPPER = None
        
        GKVariable.__init__(self, name, value, lb, ub, integer)

        

class GK_CV(GK_SV):
    """Controlled Variable. Inherits variable """
    
    def __init__(self, name='', value=0, lb=None, ub=None, gk_model=None, model_path=None, integer=False):

        # prevents the __setattr__ function from sending options to the server
        # until the __init__ function has completed since they should only be
        # sent if changed from their defaults
        self.__dict__['_initialized'] = False
        
        
        self.type = 'CV'
        
        # CV specific options
        self.BIAS = None
        self.COST = None
        self.CRITICAL = None
        self.FDELAY = 0
        self.LSTVAL = None
        self.MEAS_GAP = None
        self.PSTATUS = None
        self.SP = None
        self.SPHI = None
        self.SPLO = None
        self.STATUS = None
        self.TAU = None
        self.TIER = None
        self.TR_INIT = 0
        self.TR_OPEN = None
        self.VDVL = None
        self.VLACTION = None
        self.VLHI = None
        self.VLLO = None
        self.WMEAS = None
        self.WMODEL = None
        self.WSP = None
        self.WSPHI = None
        self.WSPLO = None

        GK_SV.__init__(self, name=name, value=value, lb=lb, ub=ub, gk_model=gk_model, model_path=model_path, integer=integer)

    def meas(self,measurement):
        self.MEAS = measurement
        #open measurement.dbs file
        f = open(os.path.join(self.path,'measurements.dbs'),'a')
        #write measurement
        f.write(self.name+'.MEAS = '+str(measurement)+', 1, none\n')
        #close file
        f.close()
        
        #write tag file
        f = open(os.path.join(self.path,self.name),'w')
        #write measurement
        f.write(str(measurement))
        #close tag file
        f.close()



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
import os

here = os.path.abspath(os.path.dirname(__file__))
# Get the long description from the README file
#with open(os.path.join(here, 'description.rst'), encoding='utf-8') as f:
#    long_description = f.read()
long_description = """GEKKO
=====

GEKKO is a python package for machine learning and optimization, specializing in
dynamic optimization of differential algebraic equations (DAE) systems. GEKKO 
provides a user-friendly interface to the powerful APMonitor optimization suite
on the back end. It is coupled with large-scale solvers for linear, quadratic,
nonlinear, and mixed integer programming (LP, QP, NLP, MILP, MINLP). Capabilities
include machine learning, discrete or continuous models, real-time optimization, 
simulation, moving horizon estimation, and nonlinear predictive control."""

def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths

#gather all files for GUI support
gui_files = package_files('gekko/static')
#add APM executable file
extra_files = gui_files + ['bin/apm.exe','bin/apm','bin/apm_arm','bin/apm_mac']

#APM binaries based on OS (currently only available for Windows)
#if os.name == 'nt':
#    apm_binary = {'gekko': ['bin/apm.exe']}
#else:
#    apm_binary = []
#elif linux:
#   apm_binary = ['gekko/bin/apmonitor' AND LA libaries]

# versions: a (alpha), b (beta), rc (release candidate)
# update version here, __init__.py, and create a GitHub release
setup(name='gekko',
    version='0.2.8',
    description='Machine learning and optimization for dynamic systems',
    long_description=long_description,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    keywords='differential deep learning reinforcement optimization mixed-integer',
    url='https://github.com/BYU-PRISM/GEKKO',
    author='BYU PRISM Lab',
    author_email='john_hedengren@byu.edu',
    license='MIT',
    packages=find_packages(),
    install_requires=[
        # 'APMonitor>=0.33',
        'flask',
        'flask_cors',
        'numpy>=1.8'#,
        #'ujson',
    ],
#    include_package_data=True,
#    package_dir={'':'gekko'},
    package_data={'gekko': extra_files},#['bin/apm.exe','static/*']},
#   TODO add testing
#    test_suite='pytest.collector',
#    tests_require=['pytest'],
    python_requires='>=2.6',
    zip_safe=False)



# Import
import string
import sys

# Get Python version
ver = sys.version_info[0]
#print('Version: '+str(ver))
if ver==2:  # Python 2
    import urllib    
else:       # Python 3+
    import urllib.request, urllib.parse, urllib.error
    #import socket

if ver==2:  # Python 2

    def cmd(server, app, aline, disp=True, debug=1):
        '''Send a request to the server \n \
           server = address of server \n \
           app      = application name \n \
           aline  = line to send to server \n \
           disp = Print output \n'''
        try:
            # Web-server URL address
            url_base = string.strip(server) + '/online/apm_line.php'
            app = app.lower()
            app.replace(" ", "")
            params = urllib.urlencode({'p': app, 'a': aline})
            f = urllib.urlopen(url_base, params)
            # initialize apm_error recording
            record_error = False
            apm_error = ''
            # Stream solution output
            if(aline=='solve'):
                line = ''
                while True:
                    char = f.read(1)
                    if not char:
                        break
                    elif char == '\n':
                        if disp: 
                            print(line)
                        if debug >= 1:
                            # Start recording output if error is detected
                            if '@error' in line:
                                record_error = True
                            if record_error:
                                apm_error+= line + '\n'
                        line = ''
                    else:
                        line += char
            # Send request to web-server
            if apm_error != '': # check if any apm errors were found
                response = apm_error
            else:
                response = f.read()
        except:
            response = 'Failed to connect to server'
        return response
    

    def get_ip(server):
        '''Get current IP address \n \
           server   = address of server'''
        # get ip address for web-address lookup
        url_base = string.strip(server) + '/ip.php'
        f = urllib.urlopen(url_base)
        ip = string.strip(f.read())
        return ip

    def get_file(server,app,filename):
        '''Retrieve any file from web-server\n \
           server   = address of server \n \
           app      = application name '''
        # Retrieve IP address
        ip = get_ip(server)
        
        # Web-server URL address
        app = app.lower()
        app.replace(" ","")
        url = string.strip(server) + '/online/' + ip + '_' + app + '/' + filename
        f = urllib.urlopen(url)
        # Send request to web-server
        file = f.read()
        # Write the file
        #fh = open(filename,'w')
        #fh.write(file.replace('\r',''))
        #fh.close()
        return (file)

else:       # Python 3+
    
    def cmd(server,app,aline, disp=True, debug=1):
        '''Send a request to the server \n \
           server = address of server \n \
           app      = application name \n \
           aline  = line to send to server \n \
           disp = Print output \n'''
        try:
            # Web-server URL address
            url_base = server.strip() + '/online/apm_line.php'
            app = app.lower()
            app.replace(" ","")
            params = urllib.parse.urlencode({'p':app,'a':aline})
            en_params = params.encode()
            f = urllib.request.urlopen(url_base,en_params)
            # initialize apm_error recording
            record_error = False
            apm_error = ''
            # Stream solution output
            if(aline=='solve'):
                line = ''
                while True:
                    en_char = f.read(1)
                    char = en_char.decode()
                    if not char:
                        break
                    elif char == '\n':
                        if disp:
                            print(line)
                        if debug >= 1:
                            # Start recording output if error is detected
                            if '@error' in line:
                                record_error = True
                            if record_error:
                                apm_error+= line + '\n'
                        line = '' # reset line
                    else:
                        line += char
            if apm_error != '': # check if any apm errors were found
                response = apm_error
            else:
                # Send request to web-server
                en_response = f.read()
                response = en_response.decode()
        except:
            response = 'Failed to connect to server'
        return response

    def get_ip(server):
        '''Get current IP address \n \
           server   = address of server'''
        # get ip address for web-address lookup
        url_base = server.strip() + '/ip.php'
        f = urllib.request.urlopen(url_base)
        fip = f.read()
        ip = fip.decode().strip()
        return ip

    def get_file(server,app,filename):
        '''Retrieve any file from web-server\n \
           server   = address of server \n \
           app      = application name '''
        # Retrieve IP address
        ip = get_ip(server)
        
        # Web-server URL address
        app = app.lower()
        app.replace(" ","")
        url = server.strip() + '/online/' + ip + '_' + app + '/' + filename
        f = urllib.request.urlopen(url)
        # Send request to web-server
        file = f.read()
        # Write the file
        #fh = open(filename,'w')
        #en_file = file.decode().replace('\r','')
        #fh.write(en_file)
        #fh.close()
        return (file)
            


